import os
import tempfile
from unittest import mock

import responses  # type: ignore

from launchable.utils.http_client import get_base_url
from tests.cli_test_case import CliTestCase


class SubsetTest(CliTestCase):
    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset(self):
        pipe = "test_1.py\ntest_2.py\ntest_3.py\ntest_4.py"
        mock_json_response = {
            "testPaths": [
                [{"type": "file", "name": "test_1.py"}],
                [{"type": "file", "name": "test_2.py"}],

            ],
            "testRunner": "file",
            "rest": [
                [{"type": "file", "name": "test_3.py"}],
                [{"type": "file", "name": "test_4.py"}],

            ],
            "subsettingId": 123,
            "summary": {
                "subset": {"duration": 10, "candidates": 3, "rate": 50},
                "rest": {"duration": 10, "candidates": 3, "rate": 50}
            },
            "isObservation": False,
        }
        responses.replace(responses.POST, "{}/intake/organizations/{}/workspaces/{}/subset".format(
            get_base_url(),
            self.organization,
            self.workspace),
            json=mock_json_response,
            status=200)

        rest = tempfile.NamedTemporaryFile(delete=False)
        result = self.cli("subset", "--target", "30%", "--session",
                          self.session, "--rest", rest.name, "file", mix_stderr=False, input=pipe)
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.stdout, "test_1.py\ntest_2.py\n")
        self.assertEqual(rest.read().decode(), os.linesep.join(["test_3.py", "test_4.py"]))
        rest.close()
        os.unlink(rest.name)

        # case: rest is empty
        mock_json_response = {
            "testPaths": [
                [{"type": "file", "name": "test_1.py"}],
                [{"type": "file", "name": "test_2.py"}],
                [{"type": "file", "name": "test_3.py"}],
                [{"type": "file", "name": "test_4.py"}],
            ],
            "testRunner": "file",
            "rest": [],
            "subsettingId": 123,
            "summary": {
                "subset": {"duration": 10, "candidates": 3, "rate": 50},
                "rest": {"duration": 10, "candidates": 3, "rate": 50}
            },
            "isObservation": False,
        }
        responses.replace(responses.POST, "{}/intake/organizations/{}/workspaces/{}/subset".format(
            get_base_url(),
            self.organization,
            self.workspace),
            json=mock_json_response,
            status=200)

        rest = tempfile.NamedTemporaryFile(delete=False)
        result = self.cli("subset", "--target", "30%", "--session",
                          self.session, "--rest", rest.name, "file", mix_stderr=False, input=pipe)
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.stdout, "test_1.py\ntest_2.py\ntest_3.py\ntest_4.py\n")
        self.assertEqual(rest.read().decode(), "")
        rest.close()
        os.unlink(rest.name)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset_with_observation_mode(self):
        pipe = "test_1.py\ntest_2.py\ntest_3.py\ntest_4.py"
        mock_json_response = {
            "testPaths": [
                [{"type": "file", "name": "test_1.py"}],
                [{"type": "file", "name": "test_2.py"}],
            ],
            "testRunner": "file",
            "rest": [
                [{"type": "file", "name": "test_3.py"}],
                [{"type": "file", "name": "test_4.py"}],

            ],
            "subsettingId": 123,
            "summary": {
                "subset": {"duration": 10, "candidates": 3, "rate": 50},
                "rest": {"duration": 10, "candidates": 3, "rate": 50}
            },
            "isObservation": True,
        }

        responses.replace(responses.POST, "{}/intake/organizations/{}/workspaces/{}/subset".format(
            get_base_url(),
            self.organization,
            self.workspace),
            json=mock_json_response,
            status=200)

        observation_mode_rest = tempfile.NamedTemporaryFile(delete=False)
        result = self.cli(
            "subset",
            "--target",
            "30%",
            "--session",
            self.session,
            "--rest",
            observation_mode_rest.name,
            "--observation",
            "file",
            input=pipe,
            mix_stderr=False)
        self.assertEqual(result.exit_code, 0)

        self.assertEqual(result.stdout, "test_1.py\ntest_2.py\ntest_3.py\ntest_4.py\n")
        self.assertEqual(observation_mode_rest.read().decode(), "")
        observation_mode_rest.close()
        os.unlink(observation_mode_rest.name)

        # case: rest is empty
        mock_json_response = {
            "testPaths": [
                [{"type": "file", "name": "test_1.py"}],
                [{"type": "file", "name": "test_2.py"}],
                [{"type": "file", "name": "test_3.py"}],
                [{"type": "file", "name": "test_4.py"}],
            ],
            "testRunner": "file",
            "rest": [],
            "subsettingId": 123,
            "summary": {
                "subset": {"duration": 10, "candidates": 3, "rate": 50},
                "rest": {"duration": 10, "candidates": 3, "rate": 50}
            },
            "isObservation": True,
        }
        responses.replace(responses.POST, "{}/intake/organizations/{}/workspaces/{}/subset".format(
            get_base_url(),
            self.organization,
            self.workspace),
            json=mock_json_response,
            status=200)

        rest = tempfile.NamedTemporaryFile(delete=False)
        result = self.cli("subset", "--target", "30%", "--session",
                          self.session, "--rest", rest.name, "--observation", "file", mix_stderr=False, input=pipe)
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.stdout, "test_1.py\ntest_2.py\ntest_3.py\ntest_4.py\n")
        self.assertEqual(rest.read().decode(), "")
        rest.close()
        os.unlink(rest.name)

    @ responses.activate
    @ mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset_with_get_tests_from_previous_full_runs(self):
        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset".format(
                get_base_url(),
                self.organization,
                self.workspace),
            json={
                "testPaths": [
                    [{"type": "file", "name": "test_aaa.py"}],
                    [{"type": "file", "name": "test_bbb.py"}],
                    [{"type": "file", "name": "test_ccc.py"}],
                ],
                "testRunner": "file",
                "rest": [
                    [{"type": "file", "name": "test_eee.py"}],
                    [{"type": "file", "name": "test_fff.py"}],
                    [{"type": "file", "name": "test_ggg.py"}],
                ],
                "subsettingId": 123,
                "summary": {
                    "subset": {"duration": 10, "candidates": 3, "rate": 50},
                    "rest": {"duration": 10, "candidates": 3, "rate": 50}
                },
            },
            status=200)

        rest = tempfile.NamedTemporaryFile(delete=False)
        result = self.cli(
            "subset",
            "--target",
            "30%",
            "--session",
            self.session,
            "--rest",
            rest.name,
            "--get-tests-from-previous-sessions",
            "file",
            mix_stderr=False)

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.stdout, "test_aaa.py\ntest_bbb.py\ntest_ccc.py\n")
        self.assertEqual(rest.read().decode(), os.linesep.join(["test_eee.py", "test_fff.py", "test_ggg.py"]))
        rest.close()
        os.unlink(rest.name)

    @ responses.activate
    @ mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset_with_output_exclusion_rules(self):
        pipe = "test_aaa.py\ntest_111.py\ntest_bbb.py\ntest_222.py\ntest_ccc.py\ntest_333.py\n"
        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset".format(
                get_base_url(),
                self.organization,
                self.workspace),
            json={
                "testPaths": [
                    [{"type": "file", "name": "test_aaa.py"}],
                    [{"type": "file", "name": "test_bbb.py"}],
                    [{"type": "file", "name": "test_ccc.py"}],
                ],
                "testRunner": "file",
                "rest": [
                    [{"type": "file", "name": "test_111.py"}],
                    [{"type": "file", "name": "test_222.py"}],
                    [{"type": "file", "name": "test_333.py"}],
                ],
                "subsettingId": 123,
                "summary": {
                    "subset": {"duration": 15, "candidates": 3, "rate": 70},
                    "rest": {"duration": 6, "candidates": 3, "rate": 30}
                },
            }, status=200)

        rest = tempfile.NamedTemporaryFile(delete=False)
        result = self.cli(
            "subset",
            "--target",
            "70%",
            "--session",
            self.session,
            "--rest",
            rest.name,
            "file",
            input=pipe,
            mix_stderr=False)

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.stdout, "test_aaa.py\ntest_bbb.py\ntest_ccc.py\n")
        self.assertEqual(rest.read().decode(), os.linesep.join(["test_111.py", "test_222.py", "test_333.py"]))
        rest.close()
        os.unlink(rest.name)

        rest = tempfile.NamedTemporaryFile(delete=False)
        result = self.cli(
            "subset",
            "--target",
            "70%",
            "--session",
            self.session,
            "--rest",
            rest.name,
            "--output-exclusion-rules",
            "file",
            input=pipe,
            mix_stderr=False)

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.stdout, "test_111.py\ntest_222.py\ntest_333.py\n")

        self.assertEqual(rest.read().decode(), os.linesep.join(["test_aaa.py", "test_bbb.py", "test_ccc.py"]))
        rest.close()
        os.unlink(rest.name)

        # case: reset is empty
        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset".format(
                get_base_url(),
                self.organization,
                self.workspace),
            json={
                "testPaths": [
                    [{"type": "file", "name": "test_aaa.py"}],
                    [{"type": "file", "name": "test_bbb.py"}],
                    [{"type": "file", "name": "test_ccc.py"}],
                    [{"type": "file", "name": "test_111.py"}],
                    [{"type": "file", "name": "test_222.py"}],
                    [{"type": "file", "name": "test_333.py"}],
                ],
                "testRunner": "file",
                "rest": [],
                "subsettingId": 123,
                "summary": {
                    "subset": {"duration": 15, "candidates": 6, "rate": 100},
                    "rest": {"duration": 0, "candidates": 0, "rate": 0}
                },
            }, status=200)

        rest = tempfile.NamedTemporaryFile(delete=False)
        result = self.cli(
            "subset",
            "--target",
            "70%",
            "--session",
            self.session,
            "--rest",
            rest.name,
            "--output-exclusion-rules",
            "file",
            input=pipe,
            mix_stderr=False)

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.stdout, "")

        self.assertEqual(rest.read().decode(), os.linesep.join(
            ["test_aaa.py", "test_bbb.py", "test_ccc.py", "test_111.py", "test_222.py", "test_333.py"]))
        rest.close()
        os.unlink(rest.name)

    @ responses.activate
    @ mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset_with_observation_and_output_exclusion_rules(self):
        pipe = "test_aaa.py\ntest_111.py\ntest_bbb.py\ntest_222.py\ntest_ccc.py\ntest_333.py\n"
        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset".format(
                get_base_url(),
                self.organization,
                self.workspace),
            json={
                "testPaths": [
                    [{"type": "file", "name": "test_aaa.py"}],
                    [{"type": "file", "name": "test_bbb.py"}],
                    [{"type": "file", "name": "test_ccc.py"}],
                ],
                "testRunner": "file",
                "rest": [
                    [{"type": "file", "name": "test_111.py"}],
                    [{"type": "file", "name": "test_222.py"}],
                    [{"type": "file", "name": "test_333.py"}],
                ],
                "subsettingId": 123,
                "summary": {
                    "subset": {"duration": 15, "candidates": 3, "rate": 70},
                    "rest": {"duration": 6, "candidates": 3, "rate": 30}
                },
                "isObservation": True,
            }, status=200)

        rest = tempfile.NamedTemporaryFile(delete=False)
        result = self.cli(
            "subset",
            "--target",
            "70%",
            "--session",
            self.session,
            "--rest",
            rest.name,
            "--output-exclusion-rules",
            "--observation",
            "file",
            input=pipe,
            mix_stderr=False)

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.stdout, "")

        self.assertEqual(rest.read().decode(), os.linesep.join(
            ["test_aaa.py", "test_bbb.py", "test_ccc.py", "test_111.py", "test_222.py", "test_333.py"]))
        rest.close()
        os.unlink(rest.name)



import gzip
from unittest import TestCase

from launchable.utils.gzipgen import compress


class GzippenTest(TestCase):
    def test_compress(self):
        """Basic sanity test of """
        encoded = b''.join(compress([b'Hello', b' ', b'world']))
        msg = gzip.decompress(encoded)
        print(msg)
        self.assertEqual(msg, b'Hello world')



import click
from junitparser import TestCase, TestSuite  # type: ignore

from launchable.testpath import TestPath

from . import launchable


def path_builder(case: TestCase, suite: TestSuite, report_file: str) -> TestPath:
    test_path = []
    if suite.name:
        test_path.append({"type": "file", "name": suite.name})

    if case.classname:
        test_path.append({"type": "class", "name": case.classname})

    if case.name:
        test_path.append({"type": "testcase", "name": case.name})

    return test_path


@click.argument('reports', required=True, nargs=-1)
@launchable.record.tests
def record_tests(client, reports):
    for r in reports:
        client.report(r)

    client.path_builder = path_builder
    client.run()


@launchable.subset
def subset(client):
    if client.base_path is None:
        raise click.BadParameter("Please specify base path")

    for line in client.stdin():
        if len(line.strip()) and not line.startswith(">"):
            client.test_path(line.rstrip("\n"))

    client.run()


split_subset = launchable.CommonSplitSubsetImpls(__name__).split_subset()



import datetime
import json
import sys
from typing import Generator

import click
import dateutil.parser

from ..commands.record.case_event import CaseEvent, CaseEventType
from ..testpath import parse_test_path, unparse_test_path
from . import launchable


@click.argument('test_path_file', required=False, type=click.File('r'))
@launchable.subset
def subset(client, test_path_file):
    """Subset tests

    TEST_PATH_FILE is a file that contains test paths (e.g.
    "file=a.py#class=classA") one per line. Lines start with a hash ('#') are
    considered as a comment and ignored.
    """

    if not client.is_get_tests_from_previous_sessions and test_path_file is None:
        raise click.BadArgumentUsage("Missing argument 'TEST_PATH_FILE'.")

    if client.is_output_exclusion_rules:
        raise click.BadArgumentUsage(
            "Don't need to use `--output-exclusion-rules` option. Please use `--rest` option and use it for exclusion"
        )

    if not client.is_get_tests_from_previous_sessions:
        tps = [s.strip() for s in test_path_file.readlines()]
        for tp_str in tps:
            if not tp_str or tp_str.startswith('#'):
                continue
            try:
                tp = parse_test_path(tp_str)
            except ValueError as e:
                sys.exit(e.args[0])
            client.test_path(tp)

    client.formatter = unparse_test_path
    client.separator = '\n'
    client.run()


split_subset = launchable.CommonSplitSubsetImpls(__name__, formatter=unparse_test_path, seperator='\n').split_subset()


@click.argument('test_result_file', required=True, type=click.Path(exists=True))
@launchable.record.tests
def record_tests(client, test_result_file):
    """Record test results

    TEST_RESULT_FILE is a file that contains a JSON document or JUnit XML file
    that describes the test results.

    ## Example JSON document

    {
      "testCases": [
         {
           "testPath": "file=a.py#class=classA",
           "duration": 42,
           "status": "TEST_PASSED",
           "stdout": "This is stdout",
           "stderr": "This is stderr",
           "createdAt": "2021-10-05T12:34:00"
         }
      ]
    }

    ## JSON schema

    {
      "$id": "https://launchableinc.com/schema/RecordTestInput",
      "$schema": "https://json-schema.org/draft/2020-12/schema",
      "title": "RecordTestInput",
      "description": "The input to record test",
      "type": "object",
      "properties": {
        "testCases": {
          "description": "Result of test cases",
          "type": "array",
          "items": {
            "type": "object",
            "properties": {
              "testPath": {
                "description": "TestPath for the test",
                "type": "string"
              },
              "duration": {
                "description": "Time taken to finish the test in seconds. If unspecified, assume 0 sec.",
                "type": "number",
                "minimum": 0
              },
              "status": {
                "description": "Test result",
                "type": "string",
                "enum": ["TEST_PASSED", "TEST_FAILED", "TEST_SKIPPED"]
              },
              "stdout": {
                "description": "Standard output of the test. If unspecified, assume empty.",
                "type": "string"
              },
              "stderr": {
                "description": "Standard error of the test. If unspecified, assume empty",
                "type": "string"
              },
              "createdAt": {
                "description": "The timestamp that the test started at. If unspecified, assume the current timestamp
                that the CLI is invoked.",
                "type": "string",
                "format": "date-time"
              }
            },
            "required": ["testPath", "status"]
          }
        }
      },
      "required": ["testCases"]
    }

    ## JUnit XML TestPath mapping

    If the file path ends with '.xml', the command parses the file as a JUnit XML file. When this mode is used the subset input
    TestPath should look like 'class={classname}#testcase={testcase}'.
    """

    def parse_json(test_result_file: str) -> Generator[CaseEventType, None, None]:
        with open(test_result_file, 'r') as f:
            doc = json.load(f)
        default_created_at = datetime.datetime.now(datetime.timezone.utc).isoformat()
        for case in doc['testCases']:
            test_path = parse_test_path(case['testPath'])
            status = case['status']
            duration_secs = case['duration'] or 0
            created_at = case['createdAt'] or default_created_at

            if status not in CaseEvent.STATUS_MAP:
                raise ValueError(
                    "The status of {} should be one of {} (was {})".format(test_path,
                                                                           list(CaseEvent.STATUS_MAP.keys()), status))
            if duration_secs < 0:
                raise ValueError("The duration of {} should be positive (was {})".format(test_path, duration_secs))
            dateutil.parser.parse(created_at)

            yield CaseEvent.create(
                test_path, duration_secs, CaseEvent.STATUS_MAP[status],
                case['stdout'], case['stderr'], created_at)

    if not test_result_file.endswith('.xml'):
        client.parse_func = parse_json

    client.report(test_result_file)
    client.run()



import re

from . import launchable


@launchable.subset
def subset(client):
    prev_cls_name = None
    for line in client.stdin():
        match = re.match(r'^INSTRUMENTATION_STATUS: class=(.+)$', line)
        if match:
            cls_name = match.group(1)
            if prev_cls_name != cls_name:
                client.test_path([{"type": "class", "name": cls_name}])
                prev_cls_name = cls_name

    client.separator = ','

    client.run()


split_subset = launchable.CommonSplitSubsetImpls(__name__, seperator=',').split_subset()
record_tests = launchable.CommonRecordTestImpls(__name__).report_files()



import datetime
import glob
import os
import re
import traceback
import xml.etree.ElementTree as ET
from http import HTTPStatus
from typing import Callable, Dict, Generator, List, Optional, Tuple, Union

import click
from dateutil.parser import parse
from junitparser import JUnitXml, JUnitXmlError, TestCase, TestSuite  # type: ignore  # noqa: F401
from more_itertools import ichunked
from tabulate import tabulate

from launchable.utils.authentication import ensure_org_workspace

from ...testpath import FilePathNormalizer, TestPathComponent, unparse_test_path
from ...utils.click import KeyValueType
from ...utils.env_keys import REPORT_ERROR_KEY
from ...utils.exceptions import InvalidJUnitXMLException
from ...utils.http_client import LaunchableClient
from ...utils.logger import Logger
from ...utils.session import parse_session, read_build
from ..helper import find_or_create_session
from .case_event import CaseEvent, CaseEventType

GROUP_NAME_RULE = re.compile("^[a-zA-Z0-9][a-zA-Z0-9_-]*$")
RESERVED_GROUP_NAMES = ["group", "groups", "nogroup", "nogroups"]


def _validate_group(ctx, param, value):
    if value is None:
        return ""

    if str(value).lower() in RESERVED_GROUP_NAMES:
        raise click.BadParameter("{} is reserved name.".format(value))

    if GROUP_NAME_RULE.match(value):
        return value
    else:
        raise click.BadParameter("group option supports only alphabet(a-z, A-Z), number(0-9), '-', and '_'")


@click.group()
@click.option(
    '--base',
    'base_path',
    help='(Advanced) base directory to make test names portable',
    type=click.Path(exists=True, file_okay=False),
    metavar="DIR",
)
@click.option(
    '--session',
    'session',
    help='Test session ID',
    type=str,
)
@click.option(
    '--build',
    'build_name',
    help='build name',
    type=str,
    metavar='BUILD_NAME',
    hidden=True,
)
@click.option(
    '--subset-id',
    'subsetting_id',
    help='subset_id',
    type=str,
)
@click.option(
    '--post-chunk',
    help='Post chunk',
    default=1000,
    type=int
)
@click.option(
    "--flavor",
    "flavor",
    help='flavors',
    cls=KeyValueType,
    multiple=True,
)
@click.option(
    "--no_base_path_inference",
    "no_base_path_inference",
    help="""Do not guess the base path to relativize the test file paths.

    By default, if the test file paths are absolute file paths, it automatically
    guesses the repository root directory and relativize the paths. With this
    option, the command doesn't do this guess work.

    If --base_path is specified, the absolute file paths are relativized to the
    specified path irrelevant to this option. Use it if the guessed base path is
    incorrect.
    """,
    is_flag=True
)
@click.option(
    '--report-paths',
    help='Instead of POSTing test results, just report test paths in the report file then quit. '
         'For diagnostics. Use with --dry-run',
    is_flag=True,
    hidden=True
)
@click.option(
    '--group',
    "group",
    help='Grouping name for test results',
    type=str,
    callback=_validate_group,
)
@click.pass_context
def tests(
    context: click.core.Context,
    base_path: str,
    session: Optional[str],
    build_name: Optional[str],
    post_chunk: int,
    subsetting_id: str,
    flavor,
    no_base_path_inference,
    report_paths,
    group,
):
    logger = Logger()

    org, workspace = ensure_org_workspace()

    test_runner = context.invoked_subcommand

    client = LaunchableClient(test_runner=test_runner, dry_run=context.obj.dry_run)

    file_path_normalizer = FilePathNormalizer(base_path, no_base_path_inference=no_base_path_inference)

    try:
        if subsetting_id:
            result = get_session_and_record_start_at_from_subsetting_id(subsetting_id, client)
            session_id = result["session"]
            record_start_at = result["start_at"]
        else:
            session_id = find_or_create_session(context, session, build_name, flavor)
            build_name = read_build()
            record_start_at = get_record_start_at(session_id, client)

        build_name, test_session_id = parse_session(session_id)
    except Exception as e:
        if os.getenv(REPORT_ERROR_KEY):
            raise e
        else:
            traceback.print_exc()
            return

    # TODO: placed here to minimize invasion in this PR to reduce the likelihood of
    # PR merge hell. This should be moved to a top-level class
    class RecordTests:
        # The most generic form of parsing, where a path to a test report
        # is turned into a generator by using CaseEvent.create()
        ParseFunc = Callable[[str], Generator[CaseEventType, None, None]]

        # A common mechanism to build ParseFunc by building JUnit XML report in-memory (or build it the usual way
        # and patch it to fix things up). This is handy as some libraries
        # produce invalid / broken JUnit reports
        JUnitXmlParseFunc = Callable[[str], ET.Element]

        @property
        def path_builder(self) -> CaseEvent.TestPathBuilder:
            """
            This function, if supplied, is used to build a test path
            that uniquely identifies a test case
            """
            return self._path_builder

        @path_builder.setter
        def path_builder(self, v: CaseEvent.TestPathBuilder):
            self._path_builder = v

        @property
        def parse_func(self) -> ParseFunc:
            return self._parse_func

        @parse_func.setter
        def parse_func(self, f: ParseFunc):
            self._parse_func = f

        # setter only property that sits on top of the parse_func property
        def set_junitxml_parse_func(self, f: JUnitXmlParseFunc):
            """
            Parse XML report file with the JUnit report file, possibly with the custom parser function 'f'
            that can be used to build JUnit ET.Element tree from scratch or do some patch up.

            If f=None, the default parse code from JUnitParser module is used.
            """

            def parse(report: str) -> Generator[CaseEventType, None, None]:
                # To understand JUnit XML format, https://llg.cubic.org/docs/junit/ is helpful
                # TODO: robustness: what's the best way to deal with broken XML
                # file, if any?
                try:
                    xml = JUnitXml.fromfile(report, f)
                except Exception as e:
                    click.echo(click.style("Warning: error reading JUnitXml file {filename}: {error}".format(
                        filename=report, error=e), fg="yellow"), err=True)
                    # `JUnitXml.fromfile()` will raise `JUnitXmlError` and other lxml related errors
                    # if the file has wrong format.
                    # https://github.com/weiwei/junitparser/blob/master/junitparser/junitparser.py#L321
                    return
                if isinstance(xml, JUnitXml):
                    testsuites = [suite for suite in xml]
                elif isinstance(xml, TestSuite):
                    testsuites = [xml]
                else:
                    raise InvalidJUnitXMLException(filename=report)

                try:
                    for suite in testsuites:
                        for case in suite:
                            yield CaseEvent.from_case_and_suite(self.path_builder, case, suite, report)
                except Exception as e:
                    click.echo(click.style("Warning: error parsing JUnitXml file {filename}: {error}".format(
                        filename=report, error=e), fg="yellow"), err=True)

            self.parse_func = parse

        junitxml_parse_func = property(None, set_junitxml_parse_func)

        @property
        def check_timestamp(self):
            return self._check_timestamp

        @check_timestamp.setter
        def check_timestamp(self, enable: bool):
            self._check_timestamp = enable

        def __init__(self, dry_run=False):
            self.reports = []
            self.skipped_reports = []
            self.path_builder = CaseEvent.default_path_builder(file_path_normalizer)
            self.junitxml_parse_func = None
            self.check_timestamp = True
            self.base_path = base_path
            self.dry_run = dry_run
            self.no_base_path_inference = no_base_path_inference

        def make_file_path_component(self, filepath) -> TestPathComponent:
            """Create a single TestPathComponent from the given file path"""
            if base_path:
                filepath = os.path.relpath(filepath, start=base_path)
            return {"type": "file", "name": filepath}

        def report(self, junit_report_file: str):
            ctime = datetime.datetime.fromtimestamp(
                os.path.getctime(junit_report_file))

            if self.check_timestamp and ctime.timestamp() < record_start_at.timestamp():
                format = "%Y-%m-%d %H:%M:%S"
                logger.warning("skip: {} is too old to report. start_record_at: {} file_created_at: {}".format(
                    junit_report_file, record_start_at.strftime(format), ctime.strftime(format)))
                self.skipped_reports.append(junit_report_file)

                return

            self.reports.append(junit_report_file)

        def scan(self, base: str, pattern: str):
            """
            Starting at the 'base' path, recursively add everything that matches the given GLOB pattern

            scan('build/test-reports', '**/*.xml')
            """
            for t in glob.iglob(os.path.join(base, pattern), recursive=True):
                self.report(t)

        def run(self):
            count = 0  # count number of test cases sent
            is_observation = False

            def testcases(reports: List[str]) -> Generator[CaseEventType, None, None]:
                exceptions = []
                for report in reports:
                    try:
                        yield from self.parse_func(report)

                    except Exception as e:
                        exceptions.append(Exception("Failed to process a report file: {}".format(report), e))

                if len(exceptions) > 0:
                    # defer XML parsing exceptions so that we can send what we
                    # can send before we bail out
                    raise Exception(exceptions)

            # generator that creates the payload incrementally
            def payload(cases: Generator[TestCase, None, None],
                        test_runner, group: str) -> Tuple[Dict[str, Union[str, List]], List[Exception]]:
                nonlocal count
                cs = []
                exs = []

                while True:
                    try:
                        cs.append(next(cases))
                    except StopIteration:
                        break
                    except Exception as ex:
                        exs.append(ex)

                count += len(cs)
                return {"events": cs, "testRunner": test_runner, "group": group}, exs

            def send(payload: Dict[str, Union[str, List]]) -> None:
                res = client.request(
                    "post", "{}/events".format(session_id), payload=payload, compress=True)

                if res.status_code == HTTPStatus.NOT_FOUND:
                    if session:
                        build, _ = parse_session(session)
                        click.echo(
                            click.style(
                                "Session {} was not found. "
                                "Make sure to run `launchable record session --build {}` "
                                "before `launchable record tests`".format(
                                    session,
                                    build),
                                'yellow'),
                            err=True)
                    elif build_name:
                        click.echo(
                            click.style(
                                "Build {} was not found. "
                                "Make sure to run `launchable record build --name {}` "
                                "before `launchable record tests`".format(
                                    build_name,
                                    build_name),
                                'yellow'),
                            err=True)

                res.raise_for_status()

            def recorded_result() -> Tuple[int, int, int, float]:
                test_count = 0
                success_count = 0
                fail_count = 0
                duration = float(0)

                for tc in testcases(self.reports):
                    test_count += 1
                    status = tc.get("status")
                    if status == 0:
                        fail_count += 1
                    elif status == 1:
                        success_count += 1
                    duration += float(tc.get("duration") or 0)  # sec

                return test_count, success_count, fail_count, duration / 60   # sec to min

            try:
                tc = testcases(self.reports)

                if report_paths:
                    # diagnostics mode to just report test paths
                    for t in tc:
                        print(unparse_test_path(t['testPath']))
                    return

                exceptions = []
                for chunk in ichunked(tc, post_chunk):
                    p, es = payload(chunk, test_runner, group)

                    send(p)
                    exceptions.extend(es)

                res = client.request("patch", "{}/close".format(session_id),
                                     payload={"metadata": get_env_values(client)})
                res.raise_for_status()
                is_observation = res.json().get("isObservation", False)

                if len(exceptions) > 0:
                    raise Exception(exceptions)

            except Exception as e:
                if os.getenv(REPORT_ERROR_KEY):
                    raise e
                else:
                    traceback.print_exc()
                    return

            if count == 0:
                if len(self.skipped_reports) != 0:
                    click.echo(click.style(
                        "{} test report(s) were skipped because they were created before this build was recorded.\n"
                        "Make sure to run your tests after you run `launchable record build`.".format(
                            len(self.skipped_reports)), 'yellow'))
                    return
                else:
                    click.echo(
                        click.style(
                            "Looks like tests didn't run? "
                            "If not, make sure the right files/directories were passed into `launchable record tests`",
                            'yellow'))
                    return

            file_count = len(self.reports)
            test_count, success_count, fail_count, duration = recorded_result()

            click.echo(
                "Launchable recorded tests for build {} (test session {}) to workspace {}/{} from {} files:".format(
                    build_name,
                    test_session_id,
                    org,
                    workspace,
                    file_count,
                ))

            if is_observation:
                click.echo("(This test session is under observation mode)")

            click.echo("")

            header = ["Files found", "Tests found", "Tests passed", "Tests failed", "Total duration (min)"]

            rows = [[file_count, test_count, success_count, fail_count, "{:0.4f}".format(duration)]]
            click.echo(tabulate(rows, header, tablefmt="github"))

            click.echo(
                "\nVisit https://app.launchableinc.com/organizations/{organization}/workspaces/"
                "{workspace}/test-sessions/{test_session_id} to view uploaded test results "
                "(or run `launchable inspect tests --test-session-id {test_session_id}`)"
                .format(
                    organization=org,
                    workspace=workspace,
                    test_session_id=test_session_id,
                ))

    context.obj = RecordTests(dry_run=context.obj.dry_run)


# if we fail to determine the timestamp of the build, we err on the side of collecting more test reports
# than no test reports, so we use the 'epoch' timestamp
INVALID_TIMESTAMP = datetime.datetime.fromtimestamp(0)


def get_record_start_at(session: Optional[str], client: LaunchableClient):
    """
    Determine the baseline timestamp to be used for up-to-date checks of report files.
    Only files newer than this timestamp will be collected.

    Based on the thinking that if a build doesn't exist tests couldn't have possibly run, we attempt
    to use the timestamp of a build, with appropriate fallback.
    """
    if session is None:
        raise click.UsageError('Either --build or --session has to be specified')

    if session:
        build_name, _ = parse_session(session)

    sub_path = "builds/{}".format(build_name)

    res = client.request("get", sub_path)
    if res.status_code != 200:
        if res.status_code == 404:
            msg = "Build {} was not found. " \
                  "Make sure to run `launchable record build --name {}` before `launchable record tests`".format(
                      build_name, build_name)
        else:
            msg = "Unable to determine the timestamp of the build {}. HTTP response code was {}".format(
                build_name,
                res.status_code)
        click.echo(click.style(msg, 'yellow'), err=True)

        # to avoid stop report command
        return INVALID_TIMESTAMP

    created_at = res.json()["createdAt"]
    Logger().debug("Build {} timestamp = {}".format(build_name, created_at))
    t = parse_launchable_timeformat(created_at)
    return t


def parse_launchable_timeformat(t: str) -> datetime.datetime:
    # e.g) "2021-04-01T09:35:47.934+00:00"
    try:
        return parse(t)
    except Exception as e:
        Logger().error("parse time error {}. time: {}".format(str(e), t))
        return INVALID_TIMESTAMP


def get_session_and_record_start_at_from_subsetting_id(subsetting_id: str, client: LaunchableClient):
    s = subsetting_id.split('/')

    # subset/{id}
    if len(s) != 2:
        raise click.UsageError('Invalid subset id. like `subset/123/slice` but got {}'.format(subsetting_id))

    res = client.request("get", subsetting_id)
    if res.status_code != 200:
        raise click.UsageError(click.style("Unable to get subset information from subset id {}".format(
            subsetting_id), 'yellow'))

    build_number = res.json()["build"]["buildNumber"]
    created_at = res.json()["build"]["createdAt"]
    test_session_id = res.json()["testSession"]["id"]

    return {
        "session": "builds/{}/test_sessions/{}".format(build_number, test_session_id),
        "start_at": parse_launchable_timeformat(created_at)
    }


def get_env_values(client: LaunchableClient) -> Dict[str, str]:
    sub_path = "slack/notification/key/list"
    res = client.request("get", sub_path=sub_path)

    metadata = {}  # type: Dict[str, str]
    if res.status_code != 200:
        return metadata

    keys = res.json().get("keys", [])
    for key in keys:
        val = os.getenv(key, "")
        metadata[key] = val

    return metadata



import json
import os
import pathlib
from copy import deepcopy
from pathlib import Path
from typing import Dict, Generator, List, Optional
from xml.etree import ElementTree as ET

import click

from launchable.testpath import FilePathNormalizer

from ..commands.record.case_event import CaseEvent, CaseEventType
from . import launchable

subset = launchable.CommonSubsetImpls(__name__).scan_files('*_feature')
split_subset = launchable.CommonSplitSubsetImpls(__name__).split_subset()


REPORT_FILE_PREFIX = "TEST-"


@click.option('--json', 'json_format', help="use JSON report format", is_flag=True)
@click.argument('reports', required=True, nargs=-1)
@launchable.record.tests
def record_tests(client, reports, json_format):
    if json_format:
        for r in reports:
            client.report(r)
        client.parse_func = JSONReportParser(client).parse_func
    else:
        report_file_and_test_file_map = {}
        _record_tests_from_xml(client, reports, report_file_and_test_file_map)

        def parse_func(report: str) -> ET.ElementTree:
            tree = ET.parse(report)
            for case in tree.findall("testcase"):
                case.attrib["file"] = str(report_file_and_test_file_map[report])

            return tree

        client.junitxml_parse_func = parse_func

    client.run()


def _record_tests_from_xml(client, reports, report_file_and_test_file_map: Dict[str, str]):
    base_path = client.base_path if client.base_path else os.getcwd()
    for report in reports:
        if REPORT_FILE_PREFIX not in report:
            click.echo("{} was load skipped because it doesn't look like a report file.".format(report), err=True)
            continue

        test_file = _find_test_file_from_report_file(base_path, report)
        if test_file:
            report_file_and_test_file_map[report] = str(test_file)
            client.report(report)
        else:
            click.echo("Cannot find test file of {}".format(report), err=True)


class JSONReportParser:
    """
        client: launchable.RecordTests
    """

    def __init__(self, client):
        self.client = client
        self.file_path_normalizer = FilePathNormalizer(base_path=client.base_path,
                                                       no_base_path_inference=client.no_base_path_inference)

    def parse_func(self, report_file: str) -> Generator[CaseEventType, None, None]:
        """
        example of JSON format report
        [
        {
            "uri": "features/foo/is_it_friday_yet.feature",
            "id": "is-it-friday-yet?",
            "keyword": "Feature",
            "name": "Is it Friday yet?",
            "description": "  Everybody wants to know when it's Friday",
            "line": 1,
            "elements": [
            {
                "id": "is-it-friday-yet?;today-is-or-is-not-friday;;2",
                "keyword": "Scenario Outline",
                "name": "Today is or is not Friday",
                "description": "",
                "line": 11,
                "type": "scenario",
                "steps": [
                {
                    "keyword": "Given ",
                    "name": "today is \"Friday\"",
                    "line": 5,
                    "match": {
                    "location": "features/step_definitions/stepdefs.rb:12"
                    },
                    "result": {
                    "status": "passed",
                    "duration": 101000
                    }
                },
                {
                    "keyword": "When ",
                    "name": "I ask whether it's Friday yet",
                    "line": 6,
                    "match": {
                    "location": "features/step_definitions/stepdefs.rb:24"
                    },
                    "result": {
                    "status": "passed",
                    "duration": 11000
                    }
                },
                {
                    "keyword": "Then ",
                    "name": "I should be told \"TGIF\"",
                    "line": 7,
                    "match": {
                    "location": "features/step_definitions/stepdefs.rb:28"
                    },
                    "result": {
                    "status": "passed",
                    "duration": 481000
                    }
                }
                ]
            }
        ]
        """
        with open(report_file, 'r') as json_file:
            try:
                data = json.load(json_file)
            except Exception as e:
                raise Exception("Can't read JSON format report file {}. Make sure to confirm report file.".format(
                    report_file)) from e

        if len(data) == 0:
            click.echo("Can't find test reports from {}. Make sure to confirm report file.".format(
                report_file), err=True)

        for d in data:
            file_name = d.get("uri", "")
            class_name = d.get("name", "")
            for element in d.get("elements", []):
                test_case = element.get("name", "")
                steps = {}
                duration = 0
                statuses = []
                stderr = []
                for step in element.get("steps", []):
                    steps[step.get("keyword", "").strip()] = step.get("name", "").strip()
                    result = step.get("result", None)

                    if result:
                        # duration's unit is nano sec
                        # ref:
                        # https://github.com/cucumber/cucumber-ruby/blob/main/lib/cucumber/formatter/json.rb#L222
                        duration = duration + result.get("duration", 0)
                        statuses.append(result.get("status"))
                        if result.get("error_message", None):
                            stderr.append(result["error_message"])

                if "failed" in statuses:
                    status = CaseEvent.TEST_FAILED
                elif "undefined" in statuses:
                    status = CaseEvent.TEST_SKIPPED
                else:
                    status = CaseEvent.TEST_PASSED

                test_path = [
                    {"type": "file", "name": pathlib.Path(self.file_path_normalizer.relativize(file_name)).as_posix()},
                    {"type": "class", "name": class_name},
                    {"type": "testcase", "name": test_case},
                ]

                for k in steps:
                    test_path.append({"type": k, "name": steps[k]})

                yield CaseEvent.create(
                    test_path, duration / 1000 / 1000 / 1000, status, None, "\n".join(stderr), None, None)


def _find_test_file_from_report_file(base_path: str, report: str) -> Optional[Path]:
    """
    Find test file from cucumber report file path format
    e.g) Test-features-foo-hoge.xml -> features/foo/hoge.feature or features/foo-hoge.feature
    """

    report_file = os.path.basename(report)
    report_file = report_file.lstrip(REPORT_FILE_PREFIX)
    report_file = os.path.splitext(report_file)[0]

    list = _create_file_candidate_list(report_file)
    for l in list:
        f = Path(base_path, l + ".feature")
        if f.exists():
            return f

    return None


def _create_file_candidate_list(file: str) -> List[str]:
    list = [""]
    for c in file:
        if c == "-":
            l = len(list)
            list += deepcopy(list)
            for i in range(l):
                list[i] += '-'
            for i in range(l, len(list)):
                list[i] += '/'
        else:
            for i in range(len(list)):
                list[i] += c

    return list



import glob
import os
import pathlib
import sys
from os.path import join
from typing import Callable, List, Optional, TextIO, Union

import click
from tabulate import tabulate

from launchable.utils.authentication import get_org_workspace
from launchable.utils.session import parse_session

from ..testpath import FilePathNormalizer, TestPath
from ..utils.click import DURATION, PERCENTAGE, DurationType, KeyValueType, PercentageType
from ..utils.env_keys import REPORT_ERROR_KEY
from ..utils.http_client import LaunchableClient
from .helper import find_or_create_session
from .test_path_writer import TestPathWriter

# TODO: rename files and function accordingly once the PR landscape


@click.group(help="Subsetting tests")
@click.option(
    '--target',
    'target',
    help='subsetting target from 0% to 100%',
    type=PERCENTAGE,
)
@click.option(
    '--time',
    'duration',
    help='subsetting by absolute time, in seconds e.g) 300, 5m',
    type=DURATION,
)
@click.option(
    '--confidence',
    'confidence',
    help='subsetting by confidence from 0% to 100%',
    type=PERCENTAGE,
)
@click.option(
    '--session',
    'session',
    help='Test session ID',
    type=str,
)
@click.option(
    '--base',
    'base_path',
    help='(Advanced) base directory to make test names portable',
    type=click.Path(exists=True, file_okay=False),
    metavar="DIR",
)
@click.option(
    '--build',
    'build_name',
    help='build name',
    type=str,
    metavar='BUILD_NAME',
    hidden=True,
)
@click.option(
    '--rest',
    'rest',
    help='output the rest of subset',
    type=str,
)
@click.option(
    "--flavor",
    "flavor",
    help='flavors',
    cls=KeyValueType,
    multiple=True,
)
@click.option(
    "--split",
    "split",
    help='split',
    is_flag=True
)
@click.option(
    "--no_base_path_inference",
    "no_base_path_inference",
    help="""Do not guess the base path to relativize the test file paths.

    By default, if the test file paths are absolute file paths, it automatically
    guesses the repository root directory and relativize the paths. With this
    option, the command doesn't do this guess work.

    If --base_path is specified, the absolute file paths are relativized to the
    specified path irrelevant to this option. Use it if the guessed base path is
    incorrect.
    """,
    is_flag=True
)
@click.option(
    "--ignore-new-tests",
    "ignore_new_tests",
    help='Ignore tests that were added recently.\n\nNOTICE: this option will ignore tests that you added just now as well',
    is_flag=True
)
@click.option(
    "--observation",
    "is_observation",
    help="enable observation mode",
    is_flag=True,
)
@click.option(
    "--get-tests-from-previous-sessions",
    "is_get_tests_from_previous_sessions",
    help="get subset list from previous full tests",
    is_flag=True,
)
@click.option(
    "--output-exclusion-rules",
    "is_output_exclusion_rules",
    help="outputs the exclude test list. Switch the subset and rest.",
    is_flag=True,
)
@click.pass_context
def subset(
    context: click.core.Context,
    target: Optional[PercentageType],
    session: Optional[str],
    base_path: Optional[str],
    build_name: Optional[str],
    rest: str,
    duration: Optional[DurationType],
    flavor: KeyValueType,
    confidence: Optional[PercentageType],
    split: bool,
    no_base_path_inference: bool,
    ignore_new_tests: bool,
    is_observation: bool,
    is_get_tests_from_previous_sessions: bool,
    is_output_exclusion_rules: bool,
):

    if is_observation and is_get_tests_from_previous_sessions:
        click.echo(
            click.style(
                "Cannot use --observation and --get-tests-from-previous-sessions options at the same time",
                fg="red"),
            err=True,
        )
        sys.exit(1)

    session_id = find_or_create_session(
        context=context,
        session=session,
        build_name=build_name,
        flavor=flavor,
        is_observation=is_observation,
    )
    file_path_normalizer = FilePathNormalizer(base_path, no_base_path_inference=no_base_path_inference)

    # TODO: placed here to minimize invasion in this PR to reduce the likelihood of
    # PR merge hell. This should be moved to a top-level class

    TestPathWriter.base_path = base_path

    class Optimize(TestPathWriter):
        # test_paths: List[TestPath]  # doesn't work with Python 3.5
        # is_get_tests_from_previous_sessions: bool

        # Where we take TestPath, we also accept a path name as a string.
        TestPathLike = Union[str, TestPath]

        # output_handler: Callable[[
        #   List[TestPathLike], List[TestPathLike]], None]
        # exclusion_output_handler: Callable[[List[TestPathLike],
        # List[TestPathLike], bool], None]]

        def __init__(self, dry_run: bool = False):
            self.rest = rest
            self.test_paths = []  # type: List
            self.output_handler = self._default_output_handler
            self.exclusion_output_handler = self._default_exclusion_output_handler
            self.is_get_tests_from_previous_sessions = is_get_tests_from_previous_sessions
            self.is_output_exclusion_rules = is_output_exclusion_rules
            super(Optimize, self).__init__(dry_run=dry_run)

        def _default_output_handler(self, output: List[TestPath], rests: List[TestPath]):
            if rest:
                self.write_file(rest, rests)

            if output:
                self.print(output)

        def _default_exclusion_output_handler(self, subset: List[TestPath], rest: List[TestPath]):
            self.output_handler(rest, subset)

        def test_path(self, path: TestPathLike):
            def rel_base_path(path):
                if isinstance(path, str):
                    return pathlib.Path(file_path_normalizer.relativize(path)).as_posix()
                else:
                    return path

            if isinstance(path, str) and any(s in path for s in ('*', "?")):
                for i in glob.iglob(path, recursive=True):
                    if os.path.isfile(i):
                        self.test_paths.append(self.to_test_path(rel_base_path(i)))
                return

            """register one test"""
            self.test_paths.append(self.to_test_path(rel_base_path(path)))

        def stdin(self) -> Union[TextIO, List]:
            # To avoid the cli continue to wait from stdin
            if is_get_tests_from_previous_sessions:
                return []

            """
            Returns sys.stdin, but after ensuring that it's connected to something reasonable.

            This prevents a typical problem where users think CLI is hanging because
            they didn't feed anything from stdin
            """
            if sys.stdin.isatty():
                click.echo(
                    click.style(
                        "Warning: this command reads from stdin but it doesn't appear to be connected to anything. "
                        "Did you forget to pipe from another command?",
                        fg='yellow'),
                    err=True)
            return sys.stdin

        @staticmethod
        def to_test_path(x: TestPathLike) -> TestPath:
            """Convert input to a TestPath"""
            if isinstance(x, str):
                # default representation for a file
                return [{'type': 'file', 'name': x}]
            else:
                return x

        def scan(self, base: str, pattern: str,
                 path_builder: Optional[Callable[[str], Union[TestPath, str, None]]] = None):
            """
            Starting at the 'base' path, recursively add everything that matches the given GLOB pattern

            scan('src/test/java', '**/*.java')

            'path_builder' is a function used to map file name into a custom test path.
            It takes a single string argument that represents the portion matched to the glob pattern,
            and its return value controls what happens to that file:
                - skip a file by returning a False-like object
                - if a str is returned, that's interpreted as a path name and
                  converted to the default test path representation. Typically, `os.path.join(base,file_name)
                - if a TestPath is returned, that's added as is
            """

            if path_builder is None:
                # default implementation of path_builder creates a file name relative to `source` so as not
                # to be affected by the path
                def default_path_builder(file_name):
                    return pathlib.Path(file_path_normalizer.relativize(join(base, file_name))).as_posix()

                path_builder = default_path_builder

            for b in glob.iglob(base):
                for t in glob.iglob(join(b, pattern), recursive=True):
                    if path_builder:
                        path = path_builder(os.path.relpath(t, b))
                    if path:
                        self.test_paths.append(self.to_test_path(path))

        def get_payload(
            self,
            session_id: str,
            target: Optional[PercentageType],
            duration: Optional[DurationType],
            test_runner: str,
        ):
            payload = {
                "testPaths": self.test_paths,
                "testRunner": test_runner,
                "session": {
                    # expecting just the last component, not the whole path
                    "id": os.path.basename(session_id)
                },
                "ignoreNewTests": ignore_new_tests,
                "getTestsFromPreviousSessions": is_get_tests_from_previous_sessions,
            }

            if target is not None:
                payload["goal"] = {
                    "type": "subset-by-percentage",
                    "percentage": target,
                }
            elif duration is not None:
                payload["goal"] = {
                    "type": "subset-by-absolute-time",
                    "duration": duration,
                }
            elif confidence is not None:
                payload["goal"] = {
                    "type": "subset-by-confidence",
                    "percentage": confidence
                }

            return payload

        def run(self):
            """called after tests are scanned to compute the optimized order"""

            # When Error occurs, return the test name as it is passed.
            original_subset = self.test_paths
            original_rests = []
            summary = {}
            subset_id = ""
            is_brainless = False
            is_observation = False

            if not session_id:
                # Session ID in --session is missing. It might be caused by
                # Launchable API errors.
                pass
            else:
                try:
                    test_runner = context.invoked_subcommand
                    client = LaunchableClient(test_runner=test_runner, dry_run=context.obj.dry_run)

                    # temporarily extend the timeout because subset API response has become slow
                    # TODO: remove this line when API response return respose
                    # within 60 sec
                    timeout = (5, 180)
                    payload = self.get_payload(session_id, target, duration, test_runner)

                    res = client.request("post", "subset", timeout=timeout, payload=payload, compress=True)

                    res.raise_for_status()

                    original_subset = res.json().get("testPaths", [])
                    original_rests = res.json().get("rest", [])
                    subset_id = res.json()["subsettingId"]
                    summary = res.json()["summary"]
                    is_brainless = res.json().get("isBrainless", False)
                    is_observation = res.json().get("isObservation", False)

                except Exception as e:
                    if os.getenv(REPORT_ERROR_KEY):
                        raise e
                    else:
                        click.echo(e, err=True)
                    click.echo(click.style(
                        "Warning: the service failed to subset. Falling back to running all tests", fg='yellow'),
                        err=True)

            if len(original_subset) == 0:
                click.echo(click.style("Error: no tests found matching the path.", 'yellow'), err=True)
                return

            if split:
                click.echo("subset/{}".format(subset_id))
            else:
                output_subset, output_rests = original_subset, original_rests

                if is_observation:
                    output_subset = output_subset + output_rests
                    output_rests = []

                if is_output_exclusion_rules:
                    self.exclusion_output_handler(output_subset, output_rests)
                else:
                    self.output_handler(output_subset, output_rests)

            # When Launchable returns an error, the cli skips showing summary
            # report
            if "subset" not in summary.keys() or "rest" not in summary.keys():
                return

            build_name, test_session_id = parse_session(session_id)
            org, workspace = get_org_workspace()

            header = ["", "Candidates",
                      "Estimated duration (%)", "Estimated duration (min)"]
            rows = [
                [
                    "Subset",
                    len(original_subset),
                    summary["subset"].get("rate", 0.0),
                    summary["subset"].get("duration", 0.0),
                ],
                [
                    "Remainder",
                    len(original_rests),
                    summary["rest"].get("rate", 0.0),
                    summary["rest"].get("duration", 0.0),
                ],
                [],
                [
                    "Total",
                    len(original_subset) + len(original_rests),
                    summary["subset"].get("rate", 0.0) + summary["rest"].get("rate", 0.0),
                    summary["subset"].get("duration", 0.0) + summary["rest"].get("duration", 0.0),
                ],
            ]

            if is_brainless:
                click.echo(
                    "Your model is currently in training", err=True)

            click.echo(
                "Launchable created subset {} for build {} (test session {}) in workspace {}/{}".format(
                    subset_id,
                    build_name,
                    test_session_id,
                    org, workspace,
                ), err=True,
            )
            if is_observation:
                click.echo(
                    "(This test session is under observation mode)",
                    err=True)

            click.echo("", err=True)
            click.echo(tabulate(rows, header, tablefmt="github"), err=True)

            click.echo(
                "\nRun `launchable inspect subset --subset-id {}` to view full subset details".format(subset_id),
                err=True)

    context.obj = Optimize(dry_run=context.obj.dry_run)



import glob
import gzip
import json
import os
from pathlib import Path
from unittest import mock

import responses  # type: ignore

from launchable.test_runners.cucumber import _create_file_candidate_list
from launchable.utils.session import write_build
from tests.cli_test_case import CliTestCase


class CucumberTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath('../data/cucumber/').resolve()

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_test(self):
        reports = []
        for f in glob.iglob(str(self.test_files_dir.joinpath("report/*.xml")), recursive=True):
            reports.append(f)

        # emulate launchable record build
        write_build(self.build_name)

        result = self.cli('record', 'tests', '--base', str(self.test_files_dir), 'cucumber', *reports)

        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(responses.calls[2].request.body).decode())
        for c in payload['events']:
            del c['created_at']

        expected = self.load_json_from_file(self.test_files_dir.joinpath('record_test_result.json'))
        self.assert_json_orderless_equal(expected, payload)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_test_from_json(self):
        reports = []
        for f in glob.iglob(str(self.test_files_dir.joinpath("report/*.json")), recursive=True):
            reports.append(f)

        # emulate launchable record build
        write_build(self.build_name)

        result = self.cli('record', 'tests', 'cucumber', "--json", *reports)

        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(responses.calls[2].request.body).decode())

        for c in payload['events']:
            del c['created_at']

        expected = self.load_json_from_file(self.test_files_dir.joinpath('record_test_json_result.json'))
        self.assert_json_orderless_equal(expected, payload)

    def test_create_file_candidate_list(self):
        self.assertCountEqual(_create_file_candidate_list("a-b"), ["a/b", "a-b"])
        self.assertCountEqual(_create_file_candidate_list("a-b-c"), ["a/b/c", "a-b/c", "a/b-c", "a-b-c"])
        self.assertCountEqual(_create_file_candidate_list("a_b_c"), ["a_b_c"])



import subprocess


def check_output(*args, **kwargs):
    """A wrapper for subprocess.check_output

    In Windows, subprocess.check_output is used internally in one of those
    dependencies. If we mock out subprocess.check_output, it also traps those
    internall calls, making tests fail. This wrapper is a point to mock only
    launchable CLI initiated calls.
    """
    return subprocess.check_output(*args, **kwargs)



import os
import sys
from http import HTTPStatus

import click
from tabulate import tabulate

from ...utils.env_keys import REPORT_ERROR_KEY
from ...utils.http_client import LaunchableClient


@click.command()
@click.option(
    '--test-session-id',
    'test_session_id',
    help='test session id',
    required=True
)
def tests(test_session_id: int):
    try:
        client = LaunchableClient()
        res = client.request(
            "get", "/test_sessions/{}/events".format(test_session_id))

        if res.status_code == HTTPStatus.NOT_FOUND:
            click.echo(click.style(
                "Test session {} not found. Check test session ID and try again.".format(test_session_id), 'yellow'),
                err=True,
            )
            sys.exit(1)

        res.raise_for_status()
        results = res.json()
    except Exception as e:
        if os.getenv(REPORT_ERROR_KEY):
            raise e
        else:
            click.echo(e, err=True)
        click.echo(click.style(
            "Warning: failed to inspect tests", fg='yellow'),
            err=True)

        return

    header = ["Test Path",
              "Duration (sec)", "Status", "Uploaded At"]

    rows = []
    for result in results:
        if result.keys() >= {"testPath"}:
            rows.append(
                [
                    "#".join([path["type"] + "=" + path["name"]
                             for path in result["testPath"] if path.keys() >= {"type", "name"}]),
                    "{:0.4f}".format(result.get("duration", 0.0)),
                    result.get("status", ""),
                    result.get("createdAt", None),
                ]
            )

    click.echo(tabulate(rows, header, tablefmt="github"))



import hashlib
from datetime import tzinfo
from typing import Dict, List, Optional

from .git_log_parser import GitCommit
from .http_client import LaunchableClient


def _sha256(s: str) -> str:
    return hashlib.sha256(s.encode('utf8')).hexdigest()


def _format_tzinfo(tz: Optional[tzinfo]) -> int:
    if not tz:
        return 0
    delta = tz.utcoffset(None)
    if not delta:
        return 0
    return round(delta.total_seconds() / 60)


def _convert_git_commit(commit: GitCommit) -> Dict:
    changed_files = []
    for changed_file in commit.changed_files:
        cf = dict()
        cf['linesAdded'] = changed_file.added
        cf['linesDeleted'] = changed_file.deleted
        cf['status'] = 'M'
        cf['path'] = changed_file.path
        cf['pathTo'] = changed_file.path
        changed_files.append(cf)
    parents = dict()
    if len(commit.parents) > 0:
        # We don't know which diff is for which parent. Use the first parent.
        parents[commit.parents[0]] = changed_files
        for parent in commit.parents[1:len(commit.parents)]:
            parents[parent] = []

    d = dict()
    d['commitHash'] = commit.commit_hash
    d['authorEmailAddress'] = _sha256(commit.author_email)
    d['authorWhen'] = round(commit.author_time.timestamp() * 1000)
    d['authorTimezoneOffset'] = _format_tzinfo(commit.author_time.tzinfo)
    d['committerEmailAddress'] = _sha256(commit.committer_email)
    d['committerWhen'] = round(commit.committer_time.timestamp() * 1000)
    d['committerTimezoneOffset'] = _format_tzinfo(commit.committer_time.tzinfo)
    d['parentHashes'] = parents
    return d


def upload_commits(commits: List[GitCommit], dry_run: bool):
    payload = {
        'commits': [_convert_git_commit(commit) for commit in commits]
    }

    client = LaunchableClient(dry_run=dry_run)
    res = client.request("post", "commits/collect", payload=payload)
    res.raise_for_status()



from pathlib import Path
from unittest import TestCase

from dateutil.parser import parse

from launchable.utils.git_log_parser import ChangedFile, GitCommit, parse_git_log


class GitLogParserTest(TestCase):
    test_file_path = Path(__file__).parent.joinpath('../data/git_log_ingest/sample.out').resolve()

    def test_parse(self):
        with self.test_file_path.open('r') as fp:
            commits = parse_git_log(fp)

        self.maxDiff = None
        self.assertEqual(commits, [
            GitCommit(
                commit_hash="1f0c18ea3df6575b4132b311d52a339af34c90ba",
                parents=["b068a8a515e6cbbb2d6673ddb2c421939bd618b7"],
                author_email="example1@example.com",
                author_time=parse("2022-09-21T15:59:21-07:00"),
                committer_email="example1@example.com",
                committer_time=parse("2022-09-21T16:34:35-07:00"),
                changed_files=[
                    ChangedFile(
                        path="launchable/commands/subset.py",
                                added=24,
                                deleted=4),
                    ChangedFile(
                        path="launchable/test_runners/gradle.py",
                        added=24,
                        deleted=0),
                    ChangedFile(
                        path="tests/test_runners/test_gradle.py",
                        added=62,
                        deleted=0),
                ],
            ),
            GitCommit(
                commit_hash="b068a8a515e6cbbb2d6673ddb2c421939bd618b7",
                parents=["cb1d1b797726fe16e661d8377bd807f2508e9df4"],
                author_email="example2@example.com",
                author_time=parse("2022-09-16T17:03:52+00:00"),
                committer_email="example3@example.com",
                committer_time=parse("2022-09-16T17:03:52+00:00"),
                changed_files=[
                    ChangedFile(path="docs/README.md", added=1, deleted=1),
                    ChangedFile(path="docs/SUMMARY.md", added=1, deleted=0),
                    ChangedFile(
                        path="docs/features/predictive-test-selection/requesting-and-running-a-subset-of-tests/README.md",
                        added=1,
                        deleted=1),
                    ChangedFile(
                        path="docs/features/predictive-test-selection/requesting-and-running-a-subset-of-tests/"
                             "zero-input-subsetting.md",
                        added=26,
                        deleted=0),
                    ChangedFile(
                        path="docs/sending-data-to-launchable/README.md",
                        added=1,
                        deleted=1),
                ],
            ),
            GitCommit(
                commit_hash="cb1d1b797726fe16e661d8377bd807f2508e9df4",
                parents=[
                    "fba956e638a7d91a6a89a40132d7d1c8f5f40d71",
                    "f22ea479d9c088d0da4109118b25c3f5b1b7b96e"
                ],
                author_email="example4@example.com",
                author_time=parse("2022-09-15T14:51:22-04:00"),
                committer_email="example5@example.com",
                committer_time=parse("2022-09-15T14:51:22-04:00"),
                changed_files=[],
            ),
            GitCommit(
                commit_hash="f22ea479d9c088d0da4109118b25c3f5b1b7b96e",
                parents=["38d92db01c0603978aab3a7bd43fbb5f5957c8f3"],
                author_email="example4@example.com",
                author_time=parse("2022-09-15T14:51:14-04:00"),
                committer_email="example5@example.com",
                committer_time=parse("2022-09-15T14:51:14-04:00"),
                changed_files=[
                    ChangedFile(
                        path="docs/features/insights/flaky-tests.md",
                                added=2,
                                deleted=2),
                ],
            ),
            GitCommit(
                commit_hash="38d92db01c0603978aab3a7bd43fbb5f5957c8f3",
                parents=["fba956e638a7d91a6a89a40132d7d1c8f5f40d71"],
                author_email="example1@example.com",
                author_time=parse("2022-09-15T11:29:45-07:00"),
                committer_email="example1@example.com",
                committer_time=parse("2022-09-15T11:29:45-07:00"),
                changed_files=[
                    ChangedFile(
                        path="docs/features/insights/flaky-tests.md",
                                added=3,
                                deleted=3),
                ],
            ),
            GitCommit(
                commit_hash="fba956e638a7d91a6a89a40132d7d1c8f5f40d71",
                parents=[
                    "a08f6ab9cb534a6818998883f8712aedddd459ae",
                    "039f5f6f7874836739ec7cbfe2bf9d0ffea0f8d3",
                ],
                author_email="example2@example.com",
                author_time=parse("2022-09-13T13:28:45-04:00"),
                committer_email="example2@example.com",
                committer_time=parse("2022-09-13T13:28:45-04:00"),
                changed_files=[],
            ),
            GitCommit(
                commit_hash="039f5f6f7874836739ec7cbfe2bf9d0ffea0f8d3",
                parents=[
                    "de1076e9a6c4efd28339496ac10906693e980ce3",
                    "8a7cb5c8cef17b9d5323c0338dce9d7b8639c85b",
                ],
                author_email="example6@example.com",
                author_time=parse("2022-09-14T02:24:32+09:00"),
                committer_email="example5@example.com",
                committer_time=parse("2022-09-14T02:24:32+09:00"),
                changed_files=[],
            ),
            GitCommit(
                commit_hash="8a7cb5c8cef17b9d5323c0338dce9d7b8639c85b",
                parents=["de1076e9a6c4efd28339496ac10906693e980ce3"],
                author_email="example6@example.com",
                author_time=parse("2022-09-13T09:43:59-07:00"),
                committer_email="example6@example.com",
                committer_time=parse("2022-09-13T09:43:59-07:00"),
                changed_files=[
                    ChangedFile(
                        path="docs/resources/integrations/maven.md",
                                added=3,
                                deleted=1),
                ],
            ),
            GitCommit(
                commit_hash="a08f6ab9cb534a6818998883f8712aedddd459ae",
                parents=["de1076e9a6c4efd28339496ac10906693e980ce3"],
                author_email="example2@example.com",
                author_time=parse("2022-09-13T14:43:41+00:00"),
                committer_email="example3@example.com",
                committer_time=parse("2022-09-13T14:43:41+00:00"),
                changed_files=[
                    ChangedFile(
                        path="docs/concepts/subset.md", added=1, deleted=1),
                ],
            ),
        ])



import gzip
import json
import os
import tempfile
from pathlib import Path
from unittest import mock

import responses  # type: ignore

from launchable.utils.http_client import get_base_url
from launchable.utils.session import read_session, write_build
from tests.cli_test_case import CliTestCase


class GoTestTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath('../data/go_test/').resolve()

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset_with_session(self):
        pipe = "TestExample1\nTestExample2\nTestExample3\nTestExample4\n" \
               "ok      github.com/launchableinc/rocket-car-gotest      0.268s"
        result = self.cli('subset', '--target', '10%', '--session', self.session, 'go-test', input=pipe)
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(responses.calls[0].request.body).decode())
        expected = self.load_json_from_file(self.test_files_dir.joinpath('subset_result.json'))
        self.assert_json_orderless_equal(expected, payload)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset_without_session(self):
        # emulate launchable record build
        write_build(self.build_name)

        pipe = "TestExample1\nTestExample2\nTestExample3\nTestExample4\n" \
               "ok      github.com/launchableinc/rocket-car-gotest      0.268s"
        result = self.cli('subset', '--target', '10%', 'go-test', input=pipe)

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(read_session(self.build_name), self.session)

        payload = json.loads(gzip.decompress(responses.calls[1].request.body).decode())
        expected = self.load_json_from_file(self.test_files_dir.joinpath('subset_result.json'))
        self.assert_json_orderless_equal(expected, payload)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_tests_with_session(self):
        result = self.cli('record', 'tests', '--session',
                          self.session, 'go-test', str(self.test_files_dir.joinpath('reportv1')) + "/")
        self.assertEqual(result.exit_code, 0)

        self.assertIn('events', responses.calls[1].request.url, 'call events API')
        payload = json.loads(gzip.decompress(responses.calls[1].request.body).decode())
        # Remove timestamp because it depends on the machine clock
        for c in payload['events']:
            del c['created_at']

        expected = self.load_json_from_file(self.test_files_dir.joinpath('record_test_result.json'))
        self.assert_json_orderless_equal(expected, payload)

        self.assertIn('close', responses.calls[3].request.url, 'call close API')

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_tests_without_session(self):
        # emulate launchable record build
        write_build(self.build_name)

        result = self.cli('record', 'tests', 'go-test', str(self.test_files_dir.joinpath('reportv1')) + "/")
        self.assertEqual(result.exit_code, 0)

        self.assertEqual(read_session(self.build_name), self.session)

        self.assertIn('events', responses.calls[2].request.url, 'call events API')
        payload = json.loads(gzip.decompress(responses.calls[2].request.body).decode())
        for c in payload['events']:
            del c['created_at']

        expected = self.load_json_from_file(self.test_files_dir.joinpath('record_test_result.json'))
        self.assert_json_orderless_equal(expected, payload)

        self.assertIn('close', responses.calls[4].request.url, 'call close API')

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_tests_v2(self):
        result = self.cli('record', 'tests', '--session',
                          self.session, 'go-test', str(self.test_files_dir.joinpath('reportv2')) + "/")
        self.assertEqual(result.exit_code, 0)

        self.assertIn('events', responses.calls[1].request.url, 'call events API')
        payload = json.loads(gzip.decompress(responses.calls[1].request.body).decode())
        # Remove timestamp because it depends on the machine clock
        for c in payload['events']:
            del c['created_at']

        expected = self.load_json_from_file(self.test_files_dir.joinpath('record_test_result.json'))
        self.assert_json_orderless_equal(expected, payload)

        self.assertIn('close', responses.calls[3].request.url, 'call close API')

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_split_subset_with_same_bin(self):
        responses.replace(
            responses.POST,
            "{}/intake/organizations/{}/workspaces/{}/subset/456/slice".format(
                get_base_url(),
                self.organization,
                self.workspace,
            ),
            json={
                'testPaths': [
                    [
                        {'type': 'class', 'name': 'rocket-car-gotest'},
                        {'type': 'testcase', 'name': 'TestExample1'},
                    ],
                    [
                        {'type': 'class', 'name': 'rocket-car-gotest'},
                        {'type': 'testcase', 'name': 'TestExample2'},
                    ],
                ],
                "rest": [],
            },
            status=200,
        )

        same_bin_file = tempfile.NamedTemporaryFile(delete=False)
        same_bin_file.write(
            b'rocket-car-gotest.TestExample1\n'
            b'rocket-car-gotest.TestExample2')
        result = self.cli(
            'split-subset',
            '--subset-id',
            'subset/456',
            '--bin',
            '1/2',
            "--same-bin",
            same_bin_file.name,
            'go-test',
        )

        self.assertEqual(result.exit_code, 0)
        self.assertEqual("^TestExample1$|^TestExample2$\n", result.output)
        same_bin_file.close()
        os.unlink(same_bin_file.name)



import os
import shutil
import tempfile
from unittest import TestCase

from launchable.utils.session import (SESSION_DIR_KEY, clean_session_files, parse_session, read_build,
                                      read_session, remove_session, write_build, write_session)


class SessionTestClass(TestCase):
    build_name = '123'
    session_id = '/intake/organizations/launchableinc/workspaces/mothership/builds/123/test_sessions/13'

    def setUp(self):
        self.dir = tempfile.mkdtemp()
        os.environ[SESSION_DIR_KEY] = self.dir

    def tearDown(self):
        clean_session_files()
        del os.environ[SESSION_DIR_KEY]
        shutil.rmtree(self.dir)

    def test_write_read_build(self):
        self.assertEqual(read_build(), None)
        write_build(self.build_name)
        self.assertEqual(read_build(), self.build_name)

    def test_write_read_remove_session(self):
        write_build(self.build_name)
        write_session(self.build_name, self.session_id)
        self.assertEqual(read_session(self.build_name), self.session_id)

        remove_session()
        self.assertEqual(read_session(self.build_name), None)

    def test_read_before_write(self):
        self.assertEqual(read_session(self.build_name), None)

    def test_parse_session(self):
        session = "builds/build-name/test_sessions/123"
        build_name, session_id = parse_session(session)
        self.assertEqual(build_name, "build-name")
        self.assertEqual(session_id, "123")

        with self.assertRaises(Exception):
            parse_session("hoge/fuga")



import gzip
import json
import os
from pathlib import Path
from unittest import mock

import responses  # type: ignore

from launchable.utils.session import read_session, write_build
from tests.cli_test_case import CliTestCase


class AdbTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath('../data/adb/').resolve()
    subset_input = """INSTRUMENTATION_STATUS: class=com.launchableinc.rocketcar.ExampleInstrumentedTest2
INSTRUMENTATION_STATUS: current=1
INSTRUMENTATION_STATUS: id=AndroidJUnitRunner
INSTRUMENTATION_STATUS: numtests=2
INSTRUMENTATION_STATUS: stream=
com.launchableinc.rocketcar.ExampleInstrumentedTest2:
INSTRUMENTATION_STATUS: test=useAppContext
INSTRUMENTATION_STATUS_CODE: 1
INSTRUMENTATION_STATUS: class=com.launchableinc.rocketcar.ExampleInstrumentedTest2
INSTRUMENTATION_STATUS: current=1
INSTRUMENTATION_STATUS: id=AndroidJUnitRunner
INSTRUMENTATION_STATUS: numtests=2
INSTRUMENTATION_STATUS: stream=.
INSTRUMENTATION_STATUS: test=useAppContext
INSTRUMENTATION_STATUS_CODE: 0
INSTRUMENTATION_STATUS: class=com.launchableinc.rocketcar.ExampleInstrumentedTest
INSTRUMENTATION_STATUS: current=2
INSTRUMENTATION_STATUS: id=AndroidJUnitRunner
INSTRUMENTATION_STATUS: numtests=2
INSTRUMENTATION_STATUS: stream=
com.launchableinc.rocketcar.ExampleInstrumentedTest:
INSTRUMENTATION_STATUS: test=useAppContext
INSTRUMENTATION_STATUS_CODE: 1
INSTRUMENTATION_STATUS: class=com.launchableinc.rocketcar.ExampleInstrumentedTest
INSTRUMENTATION_STATUS: current=2
INSTRUMENTATION_STATUS: id=AndroidJUnitRunner
INSTRUMENTATION_STATUS: numtests=2
INSTRUMENTATION_STATUS: stream=.
INSTRUMENTATION_STATUS: test=useAppContext
INSTRUMENTATION_STATUS_CODE: 0
INSTRUMENTATION_RESULT: stream=

Time: 0.011

OK (2 tests)


INSTRUMENTATION_CODE: -1
"""

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset(self):
        # emulate launchable record build
        write_build(self.build_name)

        result = self.cli('subset', '--target', '10%', 'adb', input=self.subset_input)
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(read_session(self.build_name), self.session)

        payload = json.loads(gzip.decompress(responses.calls[1].request.body).decode())
        expected = self.load_json_from_file(self.test_files_dir.joinpath('subset_result.json'))
        self.assert_json_orderless_equal(payload, expected)



import os
from typing import Tuple

import click
import requests

from .env_keys import ORGANIZATION_KEY, TOKEN_KEY, WORKSPACE_KEY


def get_org_workspace():
    token = os.getenv(TOKEN_KEY)
    if token:
        try:
            _, user, _ = token.split(":", 2)
            org, workspace = user.split("/", 1)
            return org, workspace
        except ValueError:
            return None, None

    return os.getenv(ORGANIZATION_KEY), os.getenv(WORKSPACE_KEY)


def ensure_org_workspace() -> Tuple[str, str]:
    org, workspace = get_org_workspace()
    if org is None or workspace is None:
        raise click.UsageError(
            click.style(
                "Could not identify Launchable organization/workspace. "
                "Please confirm if you set LAUNCHABLE_TOKEN or LAUNCHABLE_ORGANIZATION and "
                "LAUNCHABLE_WORKSPACE environment variables",
                fg="red"))
    return org, workspace


def authentication_headers():
    token = os.getenv(TOKEN_KEY)
    if token:
        return {'Authorization': 'Bearer {}'.format(token)}

    if os.getenv('EXPERIMENTAL_GITHUB_OIDC_TOKEN_AUTH'):
        req_url = os.getenv('ACTIONS_ID_TOKEN_REQUEST_URL')
        rt_token = os.getenv('ACTIONS_ID_TOKEN_REQUEST_TOKEN')
        if not req_url or not rt_token:
            raise click.UsageError(
                click.style(
                    "GitHub Actions OIDC tokens cannot be retrieved."
                    "Confirm that you have added necessary permissions following "
                    "https://docs.github.com/en/actions/deployment/security-hardening-your-deployments/configuring-openid-connect-in-cloud-providers#adding-permissions-settings",  # noqa: E501
                    fg="red"))
        r = requests.get(req_url,
                         headers={
                             'Authorization': 'Bearer {}'.format(rt_token),
                             'Accept': 'application/json; api-version=2.0',
                             'Content-Type': 'application/json',
                         })
        r.raise_for_status()
        return {'Authorization': 'Bearer {}'.format(r.json()['value'])}

    if os.getenv('GITHUB_ACTIONS'):
        headers = {
            'GitHub-Actions': os.environ['GITHUB_ACTIONS'],
            'GitHub-Run-Id': os.environ['GITHUB_RUN_ID'],
            'GitHub-Repository': os.environ['GITHUB_REPOSITORY'],
            'GitHub-Workflow': os.environ['GITHUB_WORKFLOW'],
            'GitHub-Run-Number': os.environ['GITHUB_RUN_NUMBER'],
            'GitHub-Event-Name': os.environ['GITHUB_EVENT_NAME'],
            'GitHub-Sha': os.environ['GITHUB_SHA'],
        }

        # GITHUB_PR_HEAD_SHA might not exist
        pr_head_sha = os.getenv('GITHUB_PR_HEAD_SHA')
        if pr_head_sha:
            headers['GitHub-Pr-Head-Sha'] = pr_head_sha

        return headers
    return {}



import click

from launchable.utils.click import GroupWithAlias

from .subset import subset
from .tests import tests


@click.group(cls=GroupWithAlias)
def inspect():
    pass


inspect.add_command(subset)
inspect.add_command(tests)



from unittest import TestCase

from launchable.utils.glob import compile


class GlobTest(TestCase):
    def check(self, glob: str, matches, not_matches):
        p = compile(glob)
        for m in matches:
            self.assertTrue(p.fullmatch(m), "%s is expected to match %s" % (glob, m))
        for m in not_matches:
            self.assertFalse(p.fullmatch(m), "%s is expected not to match %s" % (glob, m))

    def test_everything(self):
        self.check(
            glob='*.txt',
            matches=[
                "foo.txt",
                "bar.txt",
                "f.txt",
                ".txt"
            ],
            not_matches=[
                "dir/foo.txt",
                "foo.exe",
                "f_txt"
            ]
        )
        self.check(
            glob='???.txt',
            matches=[
                "foo.txt",
                "bar.txt",
            ],
            not_matches=[
                "dir/foo.txt",
                "f.txt",
                "food.txt",
                "foo.exe"
            ]
        )
        self.check(
            glob='**/*.txt',
            matches=[
                "foo.txt",
                "a/foo.txt",
                "a/b/foo.txt",
                "a\\b\\foo.txt"
            ],
            not_matches=[
                "b.exe",
                "footxt",
                "a/footxt"
            ]
        )
        self.check(
            glob='**/*$*.class',
            matches=[
                "f$f.class",
                "foo/bar$zot.class"
            ],
            not_matches=[
                "foo.class",
                "foo/bar.class"
            ]
        )



import json
import os
import shutil
import tempfile
from unittest import mock

import responses  # type: ignore

from launchable.utils.session import SESSION_DIR_KEY, clean_session_files, read_build
from tests.cli_test_case import CliTestCase


class BuildTest(CliTestCase):
    def setUp(self):
        self.dir = tempfile.mkdtemp()
        os.environ[SESSION_DIR_KEY] = self.dir

    def tearDown(self):
        clean_session_files()
        del os.environ[SESSION_DIR_KEY]
        shutil.rmtree(self.dir)

    # make sure the output of git-submodule is properly parsed
    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    @mock.patch('launchable.utils.subprocess.check_output')
    def test_submodule(self, mock_check_output):
        mock_check_output.side_effect = [
            # the first call is git rev-parse HEAD
            ('c50f5de0f06fe16afa4fd1dd615e4903e40b42a2').encode(),
            # the second call is git submodule status --recursive
            (
                ' 491e03096e2234dab9a9533da714fb6eff5dcaa7 foo (v1.51.0-560-g491e030)\n'
                ' 8bccab48338219e73c3118ad71c8c98fbd32a4be bar-zot (v1.32.0-516-g8bccab4)\n'
            ).encode()
        ]

        self.assertEqual(read_build(), None)

        result = self.cli("record", "build", "--no-commit-collection", "--name", self.build_name)
        self.assertEqual(result.exit_code, 0)

        # Name & Path should both reflect the submodule path
        self.assertTrue("| ./bar-zot | ./bar-zot | 8bccab48338219e73c3118ad71c8c98fbd32a4be |" in result.stdout)

        payload = json.loads(responses.calls[0].request.body.decode())
        self.assert_json_orderless_equal(
            {
                "buildNumber": "123",
                "commitHashes": [
                    {
                        "repositoryName": ".",
                        "commitHash": "c50f5de0f06fe16afa4fd1dd615e4903e40b42a2"
                    },
                    {
                        "repositoryName": "./foo",
                        "commitHash": "491e03096e2234dab9a9533da714fb6eff5dcaa7"
                    },
                    {
                        "repositoryName": "./bar-zot",
                        "commitHash": "8bccab48338219e73c3118ad71c8c98fbd32a4be"
                    },
                ]
            }, payload)

        self.assertEqual(read_build(), self.build_name)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    @mock.patch('launchable.utils.subprocess.check_output')
    def test_no_submodule(self, mock_check_output):
        mock_check_output.side_effect = [
            # the call is git rev-parse HEAD
            ('c50f5de0f06fe16afa4fd1dd615e4903e40b42a2').encode(),
        ]

        self.assertEqual(read_build(), None)

        result = self.cli("record", "build", "--no-commit-collection", "--no-submodules", "--name", self.build_name)
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(responses.calls[0].request.body.decode())
        self.assert_json_orderless_equal(
            {
                "buildNumber": "123",
                "commitHashes": [
                    {
                        "repositoryName": ".",
                        "commitHash": "c50f5de0f06fe16afa4fd1dd615e4903e40b42a2"
                    },
                ]
            }, payload)

        self.assertEqual(read_build(), self.build_name)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_no_git_directory(self):
        orig_dir = os.getcwd()
        try:
            os.chdir(self.dir)
            self.assertEqual(read_build(), None)

            self.cli("record", "build", "--no-commit-collection", "--commit",
                     ".=c50f5de0f06fe16afa4fd1dd615e4903e40b42a2", "--name", self.build_name)

            payload = json.loads(responses.calls[0].request.body.decode())
            self.assert_json_orderless_equal(
                {
                    "buildNumber": "123",
                    "commitHashes": [
                        {
                            "repositoryName": ".",
                            "commitHash": "c50f5de0f06fe16afa4fd1dd615e4903e40b42a2"
                        },
                    ]
                }, payload)

            self.assertEqual(read_build(), self.build_name)
        finally:
            os.chdir(orig_dir)



import os
from unittest import TestCase, mock

from launchable.utils.authentication import authentication_headers, get_org_workspace


class AuthenticationTest(TestCase):
    @mock.patch.dict(os.environ, {}, clear=True)
    def test_get_org_workspace_no_environment_variables(self):
        org, workspace = get_org_workspace()
        self.assertIsNone(org)
        self.assertIsNone(workspace)

    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": "invalid"})
    def test_get_org_workspace_invalid_LAUNCHABLE_TOKEN(self):
        org, workspace = get_org_workspace()
        self.assertIsNone(org)
        self.assertIsNone(workspace)

    @mock.patch.dict(os.environ,
                     {"LAUNCHABLE_TOKEN": "v1:launchableinc/test:token"})
    def test_get_org_workspace_valid_LAUNCHABLE_TOKEN(self):
        org, workspace = get_org_workspace()
        self.assertEqual("launchableinc", org)
        self.assertEqual("test", workspace)

    @mock.patch.dict(
        os.environ,
        {"LAUNCHABLE_ORGANIZATION": "launchableinc", "LAUNCHABLE_WORKSPACE": "test"},
        clear=True,
    )
    def test_get_org_workspace_LAUNCHABLE_ORGANIZATION_and_LAUNCHABLE_WORKSPACE(
            self):
        org, workspace = get_org_workspace()
        self.assertEqual("launchableinc", org)
        self.assertEqual("test", workspace)

    @mock.patch.dict(
        os.environ,
        {"LAUNCHABLE_TOKEN": "v1:token_org/token_wp:token",
            "LAUNCHABLE_ORGANIZATION": "org", "LAUNCHABLE_WORKSPACE": "wp"},
    )
    def test_get_org_workspace_LAUNCHABLE_TOKEN_and_LAUNCHABLE_ORGANIZATION_and_LAUNCHABLE_WORKSPACE(
            self):
        org, workspace = get_org_workspace()
        self.assertEqual("token_org", org)
        self.assertEqual("token_wp", workspace)

    @mock.patch.dict(os.environ, {}, clear=True)
    def test_authentication_headers_empty(self):
        header = authentication_headers()
        self.assertEqual(len(header), 0)

    @mock.patch.dict(os.environ,
                     {"LAUNCHABLE_TOKEN": "v1:launchableinc/test:token"})
    def test_authentication_headers_LAUNCHABLE_TOKEN(self):
        header = authentication_headers()
        self.assertEqual(len(header), 1)
        self.assertEqual(
            header["Authorization"],
            "Bearer v1:launchableinc/test:token")

    @mock.patch.dict(
        os.environ,
        {"GITHUB_ACTIONS": "true", "GITHUB_RUN_ID": "1", "GITHUB_REPOSITORY": "launchableinc/test",
         "GITHUB_WORKFLOW": "build", "GITHUB_RUN_NUMBER": "1", "GITHUB_EVENT_NAME": "push",
         "GITHUB_PR_HEAD_SHA": "test0", "GITHUB_SHA": "test1"},
        clear=True,
    )
    def test_authentication_headers_GitHub_Actions_with_PR_head(self):
        header = authentication_headers()
        self.assertEqual(len(header), 8)
        self.assertEqual(header["GitHub-Actions"], "true")
        self.assertEqual(header["GitHub-Run-Id"], "1")
        self.assertEqual(header["GitHub-Repository"], "launchableinc/test")
        self.assertEqual(header["GitHub-Workflow"], "build")
        self.assertEqual(header["GitHub-Run-Number"], "1")
        self.assertEqual(header["GitHub-Event-Name"], "push")
        self.assertEqual(header["GitHub-Pr-Head-Sha"], "test0")
        self.assertEqual(header["GitHub-Sha"], "test1")

    @mock.patch.dict(
        os.environ,
        {"GITHUB_ACTIONS": "true", "GITHUB_RUN_ID": "1", "GITHUB_REPOSITORY": "launchableinc/test",
         "GITHUB_WORKFLOW": "build", "GITHUB_RUN_NUMBER": "1", "GITHUB_EVENT_NAME": "push",
         "GITHUB_SHA": "test"},
        clear=True,
    )
    def test_authentication_headers_GitHub_Actions_without_PR_head(self):
        header = authentication_headers()
        self.assertEqual(len(header), 7)
        self.assertEqual(header["GitHub-Actions"], "true")
        self.assertEqual(header["GitHub-Run-Id"], "1")
        self.assertEqual(header["GitHub-Repository"], "launchableinc/test")
        self.assertEqual(header["GitHub-Workflow"], "build")
        self.assertEqual(header["GitHub-Run-Number"], "1")
        self.assertEqual(header["GitHub-Event-Name"], "push")
        self.assertEqual(header["GitHub-Sha"], "test")

    @mock.patch.dict(
        os.environ,
        {"LAUNCHABLE_TOKEN": "v1:launchableinc/test:token", "GITHUB_ACTIONS": "true", "GITHUB_RUN_ID": "1",
         "GITHUB_REPOSITORY": "launchableinc/test", "GITHUB_WORKFLOW": "build", "GITHUB_RUN_NUMBER": "1",
         "GITHUB_EVENT_NAME": "push", "GITHUB_SHA": "test"},
        clear=True,
    )
    def test_authentication_headers_LAUNCHABLE_TOKEN_and_GitHub_Actions(self):
        header = authentication_headers()
        self.assertEqual(len(header), 1)
        self.assertEqual(
            header["Authorization"],
            "Bearer v1:launchableinc/test:token")



import os
from unittest import mock

import responses  # type: ignore

from launchable.utils.http_client import get_base_url
from tests.cli_test_case import CliTestCase


class SubsetTest(CliTestCase):
    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset(self):
        subset_id = 456
        responses.replace(responses.GET, "{}/intake/organizations/{}/workspaces/{}/subset/{}".format(
            get_base_url(), self.organization, self.workspace, subset_id), json={
            "testPaths": [
                {"testPath": [
                    {"type": "file", "name": "test_file1.py"}], "duration": 1200},
                {"testPath": [
                    {"type": "file", "name": "test_file3.py"}], "duration": 600},
            ],
            "rest": [
                {"testPath": [
                    {"type": "file", "name": "test_file4.py"}], "duration": 1800},
                {"testPath": [
                    {"type": "file", "name": "test_file2.py"}], "duration": 100}

            ]
        }, status=200)

        result = self.cli('inspect', 'subset', '--subset-id', subset_id, mix_stderr=False)
        expect = """|   Order | Test Path          | In Subset   |   Estimated duration (sec) |
|---------|--------------------|-------------|----------------------------|
|       1 | file=test_file1.py | ✔           |                        1.2 |
|       2 | file=test_file3.py | ✔           |                        0.6 |
|       3 | file=test_file4.py |             |                        1.8 |
|       4 | file=test_file2.py |             |                        0.1 |
"""

        self.assertEqual(result.stdout, expect)



import os
import re
import sys
from typing import List

import click
from tabulate import tabulate

from ...utils import subprocess
from ...utils.authentication import get_org_workspace
from ...utils.click import KeyValueType
from ...utils.env_keys import REPORT_ERROR_KEY
from ...utils.http_client import LaunchableClient
from ...utils.session import clean_session_files, write_build
from .commit import commit


@click.command()
@click.option(
    '--name',
    'build_name',
    help='build name',
    required=True,
    type=str,
    metavar='BUILD_NAME'
)
@click.option(
    '--source',
    help='path to local Git workspace, optionally prefixed by a label.  '
    ' like --source path/to/ws or --source main=path/to/ws',
    default=["."],
    metavar="REPO_NAME",
    multiple=True
)
@click.option(
    '--max-days',
    help="the maximum number of days to collect commits retroactively",
    default=30
)
@click.option(
    '--no-submodules',
    is_flag=True,
    help="stop collecting information from Git Submodules",
    default=False
)
@click.option(
    '--no-commit-collection',
    is_flag=True,
    help="""do not collect commit data.

    This is useful if the repository is a shallow clone and the RevWalk is not
    possible. The commit data must be collected with a separate fully-cloned
    repository.
    """,
    default=False
)
@click.option('--scrub-pii', is_flag=True, help='Scrub emails and names', hidden=True)
@click.option(
    '--commit',
    'commits',
    help="set repository name and commit hash when you use --no-commit-collection option",
    multiple=True,
    default=[],
    cls=KeyValueType,
)
@click.pass_context
def build(ctx: click.core.Context, build_name: str, source: List[str], max_days: int, no_submodules: bool,
          no_commit_collection: bool, scrub_pii: bool, commits: List[str]):
    if "/" in build_name:
        sys.exit("--name must not contain a slash")
    if not no_commit_collection and len(commits) != 0:
        sys.exit("--no-commit-collection must be specified when --commit is used")

    clean_session_files(days_ago=14)

    # This command accepts REPO_NAME=REPO_DIST and REPO_DIST
    repos = [s.split('=') if re.match(r'[^=]+=[^=]+', s) else (s, s) for s in source]
    # TODO: if repo_dist is absolute path, warn the user that that's probably
    # not what they want to do

    if no_commit_collection:
        collect_commits = False
        if len(commits) == 0:
            detect_sources = True
            detect_submodules = not no_submodules
        else:
            detect_sources = False
            detect_submodules = False
    else:
        collect_commits = True
        detect_sources = True
        detect_submodules = not no_submodules

    if collect_commits:
        for (name, repo_dist) in repos:
            ctx.invoke(commit, source=repo_dist, max_days=max_days, scrub_pii=scrub_pii)
    else:
        click.echo(click.style(
            "Warning: Commit collection is turned off. The commit data must be collected separately.",
            fg='yellow'), err=True)

    sources = []
    if detect_sources:
        try:
            for name, repo_dist in repos:
                hash = subprocess.check_output("git rev-parse HEAD".split(), cwd=repo_dist).decode().replace("\n", "")

                sources.append((name, repo_dist, hash))
        except Exception as e:
            click.echo(
                click.style(
                    "Can't get commit hash. Do you run command under git-controlled directory? "
                    "If not, please set a directory use by --source option.",
                    fg='yellow'),
                err=True)
            print(e, file=sys.stderr)
            sys.exit(1)

    submodules = []
    if detect_submodules:
        for repo_name, repo_dist in repos:
            # invoke git directly because dulwich's submodule feature was
            # broken
            submodule_stdouts = subprocess.check_output("git submodule status --recursive".split(),
                                                        cwd=repo_dist).decode().splitlines()
            for submodule_stdout in submodule_stdouts:
                # the output is e.g.
                # "+bbf213437a65e82dd6dda4391ecc5d598200a6ce sub1 (heads/master)"
                matched = re.search(r"^[\+\-U ](?P<hash>[a-f0-9]{40}) (?P<name>\S+)", submodule_stdout)
                if matched:
                    commit_hash = matched.group('hash')
                    name = matched.group('name')
                    if commit_hash and name:
                        submodules.append((repo_name + "/" + name, repo_dist + "/" + name, commit_hash))

    if len(commits) != 0:
        invalid = False
        _commits = []
        # TODO: handle extraction of flavor tuple to dict in better way for
        # >=click8.0 that returns tuple of tuples as tuple of str
        if isinstance(commits[0], str):
            for c in commits:
                k, v = c.replace("(", "").replace(")", "").replace("'", "").split(",")
                _commits.append((k.strip(), v.strip()))
        else:
            _commits = commits
        for repo_name, hash in _commits:
            if not re.match("[0-9A-Fa-f]{5,40}$", hash):
                click.echo(click.style(
                    "{}'s commit hash `{}` is invalid.".format(repo_name, hash),
                    fg="yellow"),
                    err=True)
                invalid = True
            submodules.append((repo_name, "", hash))
        if invalid:
            sys.exit(1)

    # Note: currently becomes unique command args and submodules by the hash.
    # But they can be conflict between repositories.
    uniq_submodules = {commit_hash: (name, repo_dist, commit_hash)
                       for name, repo_dist, commit_hash, in sources + submodules}.values()

    try:
        commitHashes = [{
            'repositoryName': name,
            'commitHash': commit_hash
        } for name, _, commit_hash in uniq_submodules]

        if not (commitHashes[0]['repositoryName'] and commitHashes[0]['commitHash']):
            sys.exit('Please specify --source as --source .')

        payload = {
            "buildNumber": build_name,
            "commitHashes": commitHashes
        }

        client = LaunchableClient(dry_run=ctx.obj.dry_run)

        res = client.request("post", "builds", payload=payload)
        res.raise_for_status()

    except Exception as e:
        if os.getenv(REPORT_ERROR_KEY):
            raise e
        else:
            print(e)

    org, workspace = get_org_workspace()
    click.echo(
        "Launchable recorded build {} to workspace {}/{} with commits from {} {}:\n".format(
            build_name,
            org,
            workspace,
            len(uniq_submodules),
            ("repositories" if len(uniq_submodules) > 1 else "repository"),
        ),
    )

    header = ["Name", "Path", "HEAD Commit"]
    rows = [[name, repo_dist, commit_hash] for name, repo_dist, commit_hash in uniq_submodules]
    click.echo(tabulate(rows, header, tablefmt="github"))

    write_build(build_name)



import json
import os
from unittest import mock

import responses  # type: ignore

from launchable.commands.record.session import (CIRCLECI_BUILD_URL_KEY, CIRCLECI_KEY, GITHUB_ACTIONS_KEY,
                                                GITHUB_ACTIONS_REPOSITORY_KEY, GITHUB_ACTIONS_RUN_ID_KEY,
                                                GITHUB_ACTIONS_SERVER_URL_KEY, JENKINS_BUILD_URL_KEY, JENKINS_URL_KEY)
from tests.cli_test_case import CliTestCase


class SessionTest(CliTestCase):
    """
    This test needs to specify `clear=True` in mocking because the test is run on GithubActions.
    Otherwise GithubActions will export $GITHUB_* variables at runs.
    """

    @responses.activate
    @mock.patch.dict(os.environ, {
        "LAUNCHABLE_TOKEN": CliTestCase.launchable_token,
        # LANG=C.UTF-8 is needed to run CliRunner().invoke(command).
        # Generally it's provided by shell. But in this case, `clear=True`
        # removes the variable.
        'LANG': 'C.UTF-8',
    }, clear=True)
    def test_run_session_without_flavor(self):
        result = self.cli("record", "session", "--build", self.build_name)
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(responses.calls[0].request.body.decode())
        self.assert_json_orderless_equal({"flavors": {}, "isObservation": False}, payload)

    @responses.activate
    @mock.patch.dict(os.environ, {
        "LAUNCHABLE_TOKEN": CliTestCase.launchable_token,
        'LANG': 'C.UTF-8',
    }, clear=True)
    def test_run_session_with_flavor(self):
        result = self.cli("record", "session", "--build", self.build_name,
                          "--flavor", "key=value", "--flavor", "k", "v", "--flavor", "k e y = v a l u e")
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(responses.calls[0].request.body.decode())
        self.assert_json_orderless_equal({
            "flavors": {
                "key": "value",
                "k": "v",
                "k e y": "v a l u e",
            },
            "isObservation": False, }, payload)

        with self.assertRaises(ValueError):
            result = self.cli("record", "session", "--build", self.build_name, "--flavor", "only-key")
            self.assertEqual(result.exit_code, 1)
            self.assertTrue(
                "Expected key-value like --option kye=value or --option key value." in result.output)

    @responses.activate
    @mock.patch.dict(os.environ, {
        "LAUNCHABLE_TOKEN": CliTestCase.launchable_token,
        'LANG': 'C.UTF-8',
    }, clear=True)
    def test_run_session_with_observation(self):
        result = self.cli("record", "session", "--build", self.build_name, "--observation")
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(responses.calls[0].request.body.decode())
        self.assert_json_orderless_equal({"flavors": {}, "isObservation": True}, payload)

    @responses.activate
    @mock.patch.dict(os.environ, {
        "LAUNCHABLE_TOKEN": CliTestCase.launchable_token,
        JENKINS_URL_KEY: "https://jenkins.example.com/",
        JENKINS_BUILD_URL_KEY: "https://jenkins.example.com/job/launchableinc/job/example/357/",
        'LANG': 'C.UTF-8',
    }, clear=True)
    def test_run_session_with_jenkins_url(self):
        result = self.cli("record", "session", "--build", self.build_name)
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(responses.calls[0].request.body.decode())
        self.assert_json_orderless_equal(
            {
                "flavors": {},
                "isObservation": False,
                "link": {
                    "provider": "jenkins",
                    "url": "https://jenkins.example.com/job/launchableinc/job/example/357/"},
            },
            payload)

    @responses.activate
    @mock.patch.dict(os.environ, {
        "LAUNCHABLE_TOKEN": CliTestCase.launchable_token,
        GITHUB_ACTIONS_KEY: "true",
        GITHUB_ACTIONS_SERVER_URL_KEY: "https://github.com",
        GITHUB_ACTIONS_REPOSITORY_KEY: "launchableinc/example",
        GITHUB_ACTIONS_RUN_ID_KEY: "2709244304",
        'LANG': 'C.UTF-8',
    }, clear=True)
    def test_run_session_with_github_url(self):
        result = self.cli("record", "session", "--build", self.build_name)
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(responses.calls[0].request.body.decode())
        self.assert_json_orderless_equal(
            {
                "flavors": {},
                "isObservation": False,
                "link": {
                    "provider": "github-actions",
                    "url": "https://github.com/launchableinc/example/actions/runs/2709244304",
                },
            }, payload)

    @responses.activate
    @mock.patch.dict(
        os.environ,
        {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token, CIRCLECI_KEY: "true",
         CIRCLECI_BUILD_URL_KEY:
         "https://app.circleci.com/pipelines/github/launchableinc/examples/6221/workflows/990a9987-1a21-42e5-a332-89046125e5ce/jobs/7935",  # noqa: E501
         'LANG': 'C.UTF-8', },
        clear=True)
    def test_run_session_with_circleci_url(self):
        result = self.cli("record", "session", "--build", self.build_name)
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(responses.calls[0].request.body.decode())
        self.assert_json_orderless_equal(
            {
                "flavors": {},
                "isObservation": False,
                "link": {
                    "provider": "circleci",
                    "url": "https://app.circleci.com/pipelines/github/launchableinc/examples/6221/workflows/990a9987-1a21-42e5-a332-89046125e5ce/jobs/7935",  # noqa: E501
                },
            },
            payload)



import os
from typing import Dict, List, Optional

import click

from launchable.utils import glob

from . import launchable

# Surefire has the default inclusion pattern
# https://maven.apache.org/surefire/maven-surefire-plugin/test-mojo.html#includes
# and the default exclusion pattern
# https://maven.apache.org/surefire/maven-surefire-plugin/test-mojo.html#excludes
# these variables emulates those effects.
# TODO: inclusion/exclusion are user configurable patterns, so it should be user configurable
# beyond that and to fully generalize this, there's internal discussion of
# this at https://launchableinc.atlassian.net/l/c/TXDJnn09
includes = [glob.compile(x) for x in [
    # HACK: we check extensions outside the glob. We seem to allow both source
    # file enumeration and class file enumeration
    '**/Test*.*',
    '**/*Test.*',
    '**/*Tests.*',
    '**/*TestCase.*'
]]
excludes = [glob.compile(x) for x in [
    '**/*$*'
]]

# Test if a given path name is a test that Surefire recognizes


def is_file(f: str) -> bool:
    if not (f.endswith('.java') or f.endswith(".scala") or f.endswith(".kt") or f.endswith(".class")):
        return False
    for p in excludes:
        if p.fullmatch(f):
            return False
    for p in includes:
        if p.fullmatch(f):
            return True
    return False


@click.option(
    '--test-compile-created-file',
    'test_compile_created_file',
    required=False,
    multiple=True,
    type=click.Path(exists=True),
    help="Please run `mvn test-compile` command to create input file for this option",
)
@click.argument('source_roots', required=False, nargs=-1)
@launchable.subset
def subset(client, source_roots, test_compile_created_file):

    def file2class_test_path(f: str) -> List[Dict[str, str]]:
        # remove extension
        f, _ = os.path.splitext(f)

        # directory -> package name conversion
        cls_name = f.replace(os.path.sep, '.')
        return [{"type": "class", "name": cls_name}]

    def file2test(f: str) -> Optional[List]:
        if is_file(f):
            return file2class_test_path(f)
        else:
            return None

    if len(test_compile_created_file) > 0:
        for file in test_compile_created_file:
            with open(file, 'r') as f:
                lines = f.readlines()
                for l in lines:
                    # trim trailing newline
                    l = l.strip()

                    path = file2test(l)
                    if path:
                        client.test_paths.append(path)
    else:
        for root in source_roots:
            client.scan(root, '**/*', file2test)

    client.run()


split_subset = launchable.CommonSplitSubsetImpls(__name__).split_subset()
# TestNG produces surefire-reports/testng-results.xml in TestNG's native format.
# Surefire produces TEST-*.xml in JUnit format (see Surefire's StatelessXmlReporter.getReportFile)
# In addition, TestNG also produces surefire-reports/junitreports/TEST-*.xml
# (see TestNG's JUnitReportReporter.getFileName)
# And there are more test reports in play.
#
# So to collectly find tests without duplications, we need to find surefire-reports/TEST-*.xml
# not surefire-reports/**/TEST-*.xml nor surefire-reports/*.xml
record_tests = launchable.CommonRecordTestImpls(__name__).report_files(file_mask="TEST-*.xml")



import os
from typing import Dict, List

import click

from ..utils.file_name_pattern import jvm_test_pattern
from . import launchable


@click.option('--bare',
              help='outputs class names alone',
              default=False,
              is_flag=True
              )
@click.argument('source_roots', required=False, nargs=-1)
@launchable.subset
def subset(client, bare, source_roots):
    def file2test(f: str):
        if jvm_test_pattern.match(f):
            f = f[:f.rindex('.')]   # remove extension
            # directory -> package name conversion
            cls_name = f.replace(os.path.sep, '.')
            return [{"type": "class", "name": cls_name}]
        else:
            return None

    if client.is_get_tests_from_previous_sessions:
        if len(source_roots) != 0:
            click.echo(click.style(
                "Warning: SOURCE_ROOTS are ignored when --get-tests-from-previous-sessions is used", fg="yellow"),
                err=True)
            source_roots = []
    else:
        if len(source_roots) == 0:
            raise click.UsageError(click.style("Error: Missing argument 'SOURCE_ROOTS...'.", fg="red"))

    for root in source_roots:
        client.scan(root, '**/*', file2test)

    def exclusion_output_handler(subset_tests, rest_tests):
        if client.rest:
            with open(client.rest, "w+", encoding="utf-8") as fp:
                if not bare and len(rest_tests) == 0:
                    # This prevents the CLI output to be evaled as an empty
                    # string argument.
                    fp.write('-PdummyPlaceHolder')
                else:
                    fp.write(client.separator.join(client.formatter(t) for t in rest_tests))

        classes = [to_class_file(tp[0]['name']) for tp in rest_tests]
        if bare:
            click.echo(','.join(classes))
        else:
            click.echo('-PexcludeTests=' + (','.join(classes)))
    client.exclusion_output_handler = exclusion_output_handler

    if bare:
        client.formatter = lambda x: x[0]['name']
    else:
        client.formatter = lambda x: "--tests {}".format(x[0]['name'])
        client.separator = ' '

    client.run()


@click.option('--bare',
              help='outputs class names alone',
              default=False,
              is_flag=True
              )
@launchable.split_subset
def split_subset(client, bare):
    if bare:
        client.formatter = lambda x: x[0]['name']
    else:
        client.formatter = lambda x: "--tests {}".format(x[0]['name'])
        client.separator = ' '

    def format_same_bin(s: str) -> List[Dict[str, str]]:
        return [{"type": "class", "name": s}]

    client.same_bin_formatter = format_same_bin

    client.run()


def to_class_file(class_name: str):
    return class_name.replace('.', '/') + '.class'


record_tests = launchable.CommonRecordTestImpls(__name__).report_files()



import gzip
import json
import os
import platform
from typing import Dict, Optional, Tuple

from requests import Session
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry  # type: ignore

from launchable.version import __version__

from .authentication import authentication_headers, get_org_workspace
from .env_keys import BASE_URL_KEY
from .logger import AUDIT_LOG_FORMAT, Logger

DEFAULT_BASE_URL = "https://api.mercury.launchableinc.com"


def get_base_url():
    return os.getenv(BASE_URL_KEY) or DEFAULT_BASE_URL


class DryRunResponse:
    def __init__(self, status_code, payload):
        self.status_code = status_code
        self.payload = payload

    def raise_for_status(self):
        return

    def json(self):
        return self.payload


class LaunchableClient:
    def __init__(self, base_url: str = "", session: Optional[Session] = None,
                 test_runner: Optional[str] = "", dry_run: bool = False):
        self.base_url = base_url or get_base_url()
        self.dry_run = dry_run

        if session is None:
            strategy = Retry(
                total=3,
                allowed_methods=["GET", "PUT", "PATCH", "DELETE"],
                status_forcelist=[429, 500, 502, 503, 504],
                backoff_factor=2
            )

            adapter = HTTPAdapter(max_retries=strategy)
            s = Session()
            s.mount("http://", adapter)
            s.mount("https://", adapter)
            self.session = s
        else:
            self.session = session  # type: ignore

        self.test_runner = test_runner

        self.organization, self.workspace = get_org_workspace()
        if self.organization is None or self.workspace is None:
            raise ValueError(
                "Could not identify a Launchable organization/workspace. "
                "Confirm that you set LAUNCHABLE_TOKEN "
                "(or LAUNCHABLE_ORGANIZATION and LAUNCHABLE_WORKSPACE) environment variable(s)\n"
                "See https://docs.launchableinc.com/getting-started#setting-your-api-key")

    def request(
        self,
        method: str,
        sub_path: str,
        payload: Optional[Dict] = None,
        params: Optional[Dict] = None,
        timeout: Tuple[int, int] = (5, 60),
        compress: bool = False,
    ):
        url = _join_paths(
            self.base_url, "/intake/organizations/{}/workspaces/{}".format(self.organization, self.workspace),
            sub_path)

        headers = self._headers(compress)

        Logger().audit(AUDIT_LOG_FORMAT.format("(DRY RUN) " if self.dry_run else "", method, url, headers, payload))

        if self.dry_run and method.upper() not in ["HEAD", "GET"]:
            return DryRunResponse(status_code=200, payload={
                "id": "dry-run-session",  # `record session` use this
                "testPaths": [],  # `split_subset` use this
                "rest": [],  # `split_subset` use this
            })

        data = _build_data(payload, compress)

        try:
            response = self.session.request(method, url, headers=headers, timeout=timeout, data=data, params=params)
            Logger().debug(
                "received response status:{} message:{} headers:{}".format(response.status_code, response.reason,
                                                                           response.headers)
            )
            return response
        except Exception as e:
            raise Exception("unable to post to %s" % url) from e

    def _headers(self, compress):
        h = {
            "User-Agent": "Launchable/{} (Python {}, {})".format(
                __version__,
                platform.python_version(),
                platform.platform(),
            ),
            "Content-Type": "application/json"
        }

        if compress:
            h["Content-Encoding"] = "gzip"

        if self.test_runner != "":
            h["User-Agent"] = h["User-Agent"] + " TestRunner/{}".format(self.test_runner)

        return {**h, **authentication_headers()}


def _build_data(payload, compress):
    if payload is None:
        return None

    encoded = json.dumps(payload).encode()
    if compress:
        return gzip.compress(encoded)

    return encoded


def _join_paths(*components):
    return '/'.join([c.strip('/') for c in components])



import gzip
import json
import os
from pathlib import Path
from unittest import mock

import responses  # type: ignore

from launchable.utils.http_client import get_base_url
from launchable.utils.session import write_build
from tests.cli_test_case import CliTestCase
from tests.helper import ignore_warnings


class JestTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath('../data/jest/').resolve()
    # This string generate absolute paths because the CLI requires exsisting
    # directory path
    subset_input = """
> launchable@0.1.0 test
> jest "--listTests"

{}/__tests__/pages/organizations/workspaces/index.test.tsx
{}/components/layouts/modal/snapshot.test.tsx
{}/components/molecules/mobile-sidebar/snapshot.test.tsx
{}/components/atoms/button/index.test.tsx
{}/components/layouts/sidebar/snapshot.test.tsx
{}/components/molecules/sidebar/snapshot.test.tsx
{}/components/atoms/toggle/snapshot.test.tsx
{}/components/layouts/error/snapshot.test.tsx
{}/components/layouts/email-required/snapshot.test.tsx
{}/components/layouts/loading/snapshot.test.tsx
""".format(*(os.getcwd() for _ in range(10)))

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset(self):
        # emulate launchable record build
        write_build(self.build_name)

        result = self.cli('subset', '--target', '10%', '--base',
                          os.getcwd(), 'jest', input=self.subset_input)
        print(result.output)
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(responses.calls[1].request.body).decode())
        expected = self.load_json_from_file(self.test_files_dir.joinpath('subset_result.json'))
        self.assert_json_orderless_equal(payload, expected)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    @ignore_warnings
    def test_subset_split(self):
        test_path = Path("{}/components/layouts/modal/snapshot.test.tsx".format(os.getcwd()))
        responses.replace(responses.POST,
                          "{}/intake/organizations/{}/workspaces/{}/subset".format(get_base_url(),
                                                                                   self.organization,
                                                                                   self.workspace),
                          json={'testPaths': [[{'name': str(test_path)}]],
                                'rest': [],
                                'subsettingId': 123,
                                'summary': {'subset': {'duration': 10,
                                                       'candidates': 1,
                                                       'rate': 100},
                                            'rest': {'duration': 0,
                                                     'candidates': 0,
                                                     'rate': 0}},
                                "isBrainless": False,
                                },
                          status=200)

        # emulate launchable record build
        write_build(self.build_name)

        result = self.cli('subset', '--target', '20%', '--base', os.getcwd(), '--split',
                          'jest', input=self.subset_input)

        self.assertEqual(result.exit_code, 0)
        self.assertIn('subset/123', result.output)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_test(self):
        # emulate launchable record build
        write_build(self.build_name)

        result = self.cli('record', 'tests', 'jest', str(self.test_files_dir.joinpath("junit.xml")))
        self.assertEqual(result.exit_code, 0)

        payload = json.loads(gzip.decompress(responses.calls[2].request.body).decode())
        expected = self.load_json_from_file(self.test_files_dir.joinpath('record_test_result.json'))
        self.assert_json_orderless_equal(expected, payload)



import os.path
import pathlib
import subprocess
import sys
import tempfile
import unittest

from launchable.testpath import FilePathNormalizer, parse_test_path, unparse_test_path


class TestPathEncodingTest(unittest.TestCase):
    def test_parse(self):
        self.assertEqual(parse_test_path('a=b#c=d'), [{'type': 'a', 'name': 'b'}, {'type': 'c', 'name': 'd'}])
        self.assertEqual(parse_test_path('&name=b#&type=c'), [{'name': 'b'}, {'type': 'c'}])
        self.assertEqual(parse_test_path('t=n&a=aa&x=xx'), [{'a': 'aa', 'name': 'n', 'type': 't', 'x': 'xx'}])
        self.assertEqual(parse_test_path('&a=aa&type=t&x=xx'), [{'a': 'aa', 'type': 't', 'x': 'xx'}])
        self.assertEqual(parse_test_path('a=x%25%3D%23%26'), [{'type': 'a', 'name': 'x%=#&'}])
        self.assertEqual(parse_test_path(''), [])
        # See the comment on the parse_test_path.
        # self.assertEqual(
        #     parse_test_path('&'),
        #     [{None: None}])
        # self.assertEqual(
        #     parse_test_path('&#&'),
        #     [{None: None}, {None: None}])

    def test_unparse(self):
        self.assertEqual(
            unparse_test_path([{'type': 'a', 'name': 'b'}, {'type': 'c', 'name': 'd'}]), 'a=b#c=d')
        self.assertEqual(
            unparse_test_path([{'type': None, 'name': 'b'}, {'type': 'c', 'name': None}]), '&name=b#&type=c')
        self.assertEqual(
            unparse_test_path([{'a': 'aa', 'name': 'n', 'type': 't', 'x': 'xx'}]), 't=n&a=aa&x=xx')
        self.assertEqual(
            unparse_test_path([{'a': 'aa', 'type': 't', 'x': 'xx'}]), '&a=aa&type=t&x=xx')
        self.assertEqual(unparse_test_path([{'a': 'x%=#&'}]), '&a=x%25%3D%23%26')
        self.assertEqual(unparse_test_path([]), '')
        self.assertEqual(unparse_test_path([{None: None}]), '&')
        self.assertEqual(unparse_test_path([{None: None}, {None: None}]), '&#&')


class TestFilePathNormalizer(unittest.TestCase):
    def test_relative_path(self):
        n = FilePathNormalizer()
        relpath = os.path.join('foo', 'bar', 'baz')
        self.assertEqual(relpath, n.relativize(relpath))

    def test_base_path(self):
        base = os.path.abspath('base')
        relpath = os.path.join('foo', 'bar', 'baz')
        abspath = os.path.join(base, 'foo', 'bar', 'baz')

        n = FilePathNormalizer(base_path=base)
        self.assertEqual(relpath, n.relativize(relpath))
        self.assertEqual(relpath, n.relativize(abspath))

    def test_normalize_path(self):
        relpath = os.path.join('foo', 'bar', 'baz')
        non_normalized = os.path.join('foo', 'bar', 'omit', '..', 'baz')

        n = FilePathNormalizer()
        self.assertEqual(relpath, n.relativize(non_normalized))

    def test_no_base_path_inference(self):
        base = os.path.abspath('base')
        relpath = os.path.join('foo', 'bar', 'baz')
        abspath = os.path.join(base, 'foo', 'bar', 'baz')

        n = FilePathNormalizer(no_base_path_inference=True)
        self.assertEqual(relpath, n.relativize(relpath))
        self.assertEqual(abspath, n.relativize(abspath))

        n = FilePathNormalizer(base_path=base, no_base_path_inference=True)
        self.assertEqual(relpath, n.relativize(relpath))
        self.assertEqual(relpath, n.relativize(abspath))

    @unittest.skipIf(
        sys.platform.startswith("win"),
        "tempfile creates 8.3 filenames, and it's hard to deal with them. "
        "Practically, we don't see them often, so do not support them now until it's needed."
    )
    def test_inference_git(self):
        with tempfile.TemporaryDirectory() as tempdirname:
            temppath = pathlib.PurePath(tempdirname)
            base = str(temppath.joinpath("gitrepo"))
            relpath = os.path.join('foo', 'bar', 'baz')
            abspath = os.path.join(base, 'foo', 'bar', 'baz')

            self._run_command(['git', 'init', base])

            n = FilePathNormalizer()
            self.assertEqual(relpath, n.relativize(abspath))

    @unittest.skipIf(
        sys.platform.startswith("win"),
        "tempfile creates 8.3 filenames, and it's hard to deal with them. "
        "Practically, we don't see them often, so do not support them now until it's needed. "
        "Also when this runs on Windows, GIT_AUTHOR_NAME etc. is ignored and fails."
    )
    def test_inference_git_submodule(self):
        with tempfile.TemporaryDirectory() as tempdirname:
            temppath = pathlib.PurePath(tempdirname)

            self._run_command(['git', 'init', str(temppath.joinpath("submod"))])
            self._run_command(
                ['git', 'commit', '--allow-empty', '--message', 'test commit'],
                cwd=str(temppath.joinpath("submod")))

            self._run_command(['git', 'init', str(temppath.joinpath("gitrepo"))])
            self._run_command(
                ['git', '-c', 'protocol.file.allow=always', 'submodule', 'add', str(temppath.joinpath("submod")), 'submod'],
                cwd=str(temppath.joinpath("gitrepo")))

            base = str(temppath.joinpath("gitrepo"))
            relpath = os.path.join('submod', 'foo', 'bar', 'baz')
            abspath = os.path.join(base, 'submod', 'foo', 'bar', 'baz')

            n = FilePathNormalizer()
            self.assertEqual(relpath, n.relativize(abspath))

    def _run_command(self, args, cwd=None):
        # Use check_output here to capture stderr for a better error message.
        try:
            subprocess.check_output(
                args,
                cwd=cwd,
                stderr=subprocess.PIPE,
                universal_newlines=True,
                env={
                    "GIT_AUTHOR_NAME": "Test User",
                    "GIT_AUTHOR_EMAIL": "user@example.com",
                    "GIT_COMMITTER_NAME": "Test User",
                    "GIT_COMMITTER_EMAIL": "user@example.com",
                })
        except subprocess.CalledProcessError as e:
            self.fail("Failed to execute a command: {}\nSTDOUT: {}\nSTDERR: {}\n".                format(e, e.stdout, e.stderr))


if __name__ == '__main__':
    unittest.main()



import json
import os
from pathlib import Path
from typing import Optional

from .exceptions import ParseSessionException

SESSION_DIR_KEY = 'LAUNCHABLE_SESSION_DIR'


def _session_file_dir() -> Path:
    return Path(os.environ.get(SESSION_DIR_KEY) or os.getcwd()).expanduser()


def _session_file_path() -> Path:
    return _session_file_dir() / ".launchable"


def read_build() -> Optional[str]:
    f = _session_file_path()
    try:
        if not f.exists():
            return None

        with open(str(_session_file_path())) as session_file:
            session = json.load(session_file)
            return session.get("build")

    except Exception as e:
        raise Exception("Can't read {}".format(f)) from e


def read_session(build_name: str) -> Optional[str]:
    f = _session_file_path()
    try:
        if not f.exists():
            return None

        with open(str(_session_file_path())) as session_file:
            session = json.load(session_file)
            if build_name != session.get('build', None):
                raise Exception("Build name is different from saved. input:{} saved:{}".format(
                    build_name, session.get('build', None)))

            return session.get("session")

    except Exception as e:
        raise Exception("Can't read {}".format(f)) from e


def write_build(build_name: str) -> None:
    try:
        if not _session_file_dir().exists():
            _session_file_dir().mkdir(parents=True, exist_ok=True)

        session = {}
        session["build"] = build_name

        with open(str(_session_file_path()), 'w') as session_file:
            json.dump(session, session_file)

    except Exception as e:
        raise Exception(
            "Can't write to {}. Is the path writable? "
            "If not, set the {} environment variable to specify an alternative directory for this file.".format(
                _session_file_path(),
                SESSION_DIR_KEY)) from e


def write_session(build_name: str, session_id: str) -> None:
    try:
        session = {}
        session["build"] = build_name
        session["session"] = session_id

        with open(str(_session_file_path()), 'w') as session_file:
            json.dump(session, session_file)

    except Exception as e:
        raise Exception(
            "Can't write to {}. "
            "Perhaps set the {} environment variable to specify an alternative writable path?".format(
                _session_file_path(), SESSION_DIR_KEY)) from e


def remove_session() -> None:
    """
    Call it after closing a session
    """
    if _session_file_path().exists():
        _session_file_path().unlink()


def clean_session_files(days_ago: int = 0) -> None:
    """
    Call it each build start
    """
    remove_session()


def parse_session(session: str):
    # session format:
    # builds/<build name>/test_sessions/<test session id>
    if session.count("/") != 3:
        raise ParseSessionException(session=session)

    _, build_name, _, session_id = session.split("/")
    return build_name, session_id



from xmlrunner.runner import XMLTestProgram, XMLTestRunner

"""
This project use xmlrunner command to publish xml test report.
However, the xmlrunner adds a date time suffix to test suite automatically.
Although, there are the same test cases, the Launchable realizes the differect test cases.
So I add this wrapper command to disable suffix.
"""


class LaunchableTestRunner(XMLTestRunner):
    def __init__(self, **kwargs):
        super().__init__(output='test-results', outsuffix="", **kwargs)


class LaunchableTestProgram(XMLTestProgram):
    def __init__(self, *args, **kwargs):
        kwargs["testRunner"] = LaunchableTestRunner
        super().__init__(*args, **kwargs)


LaunchableTestProgram(module=None)



import os
from unittest import mock

import responses  # type: ignore

from launchable.utils.http_client import get_base_url
from tests.cli_test_case import CliTestCase


class SubsetTest(CliTestCase):
    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset(self):
        test_session_id = 16
        responses.replace(responses.GET, "{}/intake/organizations/{}/workspaces/{}/test_sessions/{}/events".format(
            get_base_url(),
            self.organization,
            self.workspace,
            test_session_id),
            json=[
            {"testPath": [
                {
                    "type": "file",
                    "name": "test_file1.py",
                },
            ],
                "duration": 1.2,
                "stderr": "",
                "stdout": "",
                "createdAt": "2021-01-02T03:04:05.000+00:00",
                "status": "SUCCESS",
            },
            {"testPath": [
                {
                    "type": "file",
                    "name": "test_file3.py",
                },
            ],
                "duration": 0.6,
                "stderr": "",
                "stdout": "",
                "createdAt": "2021-01-02T03:04:05.000+00:00",
                "status": "SUCCESS",
            },


            {"testPath": [
                {
                    "type": "file",
                    "name": "test_file4.py",
                },
            ],
                "duration": 1.8,
                "stderr": "",
                "stdout": "",
                "createdAt": "2021-01-02T03:04:05.000+00:00",
                "status": "FAILURE",
            },
            {"testPath": [
                {
                    "type": "file",
                    "name": "test_file2.py",
                },
            ],
                "duration": 0.1,
                "stderr": "",
                "stdout": "",
                "createdAt": "2021-01-02T03:04:05.000+00:00",
                "status": "FAILURE",
            },
        ], status=200)

        result = self.cli('inspect', 'tests', '--test-session-id', test_session_id, mix_stderr=False)
        expect = """| Test Path          |   Duration (sec) | Status   | Uploaded At                   |
|--------------------|------------------|----------|-------------------------------|
| file=test_file1.py |              1.2 | SUCCESS  | 2021-01-02T03:04:05.000+00:00 |
| file=test_file3.py |              0.6 | SUCCESS  | 2021-01-02T03:04:05.000+00:00 |
| file=test_file4.py |              1.8 | FAILURE  | 2021-01-02T03:04:05.000+00:00 |
| file=test_file2.py |              0.1 | FAILURE  | 2021-01-02T03:04:05.000+00:00 |
"""

        self.assertEqual(result.stdout, expect)



import importlib
import importlib.util
import logging
from glob import glob
from os.path import basename, dirname, join

import click

from .commands.inspect import inspect
from .commands.record import record
from .commands.split_subset import split_subset
from .commands.stats import stats
from .commands.subset import subset
from .commands.verify import verify
from .utils import logger
from .version import __version__


class AppBase(object):
    def __init__(self, dry_run: bool):
        self.dry_run = dry_run


@click.group()
@click.version_option(version=__version__, prog_name='launchable-cli')
@click.option(
    '--log-level',
    'log_level',
    help='Set logger\'s log level (CRITICAL, ERROR, WARNING, AUDIT, INFO, DEBUG).',
    type=str,
    default=logger.LOG_LEVEL_DEFAULT_STR,
)
@click.option(
    '--plugins',
    'plugin_dir',
    help='Directory to load plugins from',
    type=click.Path(exists=True, file_okay=False)
)
@click.option(
    '--dry-run',
    'dry_run',
    help='Dry-run mode. No data is sent to the server. However, sometimes '
         'GET requests without payload data or side effects could be sent.'
         'note: Since the dry run log is output together with the AUDIT log, '
         'even if the log-level is set to warning or higher, the log level will '
         'be forced to be set to AUDIT.',
    is_flag=True,
)
@click.pass_context
def main(ctx, log_level, plugin_dir, dry_run):
    level = logger.get_log_level(log_level)
    # In the case of dry-run, it is forced to set the level below the AUDIT.
    # This is because the dry-run log will be output along with the audit log.
    if dry_run and level > logger.LOG_LEVEL_AUDIT:
        level = logger.LOG_LEVEL_AUDIT

    logging.basicConfig(level=level)

    # load all test runners
    for f in glob(join(dirname(__file__), 'test_runners', "*.py")):
        f = basename(f)[:-3]
        if f == '__init__':
            continue
        importlib.import_module('launchable.test_runners.%s' % f)

    # load all plugins
    if plugin_dir:
        for f in glob(join(plugin_dir, '*.py')):
            spec = importlib.util.spec_from_file_location(
                "launchable.plugins.{}".format(basename(f)[:-3]), f)
            plugin = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(plugin)

    ctx.obj = AppBase(dry_run=dry_run)


main.add_command(record)
main.add_command(subset)
main.add_command(split_subset)
main.add_command(verify)
main.add_command(inspect)
main.add_command(stats)

if __name__ == '__main__':
    main()



from typing import Optional

import click

from ..utils.http_client import LaunchableClient
from ..utils.session import read_build, read_session


def find_or_create_session(
    context: click.core.Context,
    session: Optional[str],
    build_name: Optional[str],
    flavor=[],
    is_observation: bool = False,
) -> Optional[str]:
    """Determine the test session ID to be used.

    1. If the user explicitly provides the session id via the `--session` option
    2. If the user gives no options, the current session ID is read from the session file tied to $PWD,
       or one is created from the current build name. See https://github.com/launchableinc/cli/pull/342
    3. The `--build` option is legacy compatible behaviour, in which case a session gets created and tied
       to the build.

    Args:
        session: The --session option value
        build_name: The --build option value
    """
    from .record.session import session as session_command

    if session:
        _check_observation_mode_status(session, is_observation)
        return session

    saved_build_name = read_build()
    if not saved_build_name:
        raise click.UsageError(
            click.style(
                "No saved build name found.\n"
                "To fix this, run `launchable record build`.\n"
                "If you already ran this command on a different machine, use the --session option. "
                "See https://docs.launchableinc.com/sending-data-to-launchable/managing-complex-test-session-layouts",
                fg="yellow"))

    else:
        if build_name and saved_build_name != build_name:
            raise click.UsageError(
                click.style(
                    "The build name you provided ({}) is different from the last build name recorded on this machine ({}).\n"
                    "Make sure to run `launchable record build --name {}` before you run this command.\n"
                    "If you already recorded this build on a different machine, use the --session option instead of --build. "
                    "See https://docs.launchableinc.com/sending-data-to-launchable/managing-complex-test-session-layouts".format(
                        build_name, saved_build_name, build_name), fg="yellow", ))

        session_id = read_session(saved_build_name)
        if session_id:
            _check_observation_mode_status(session_id, is_observation)
            return session_id
        else:
            context.invoke(
                session_command,
                build_name=saved_build_name,
                save_session_file=True,
                print_session=False,
                flavor=flavor,
                is_observation=is_observation,
            )
            return read_session(saved_build_name)


def _check_observation_mode_status(session: str, is_observation: bool):
    if not is_observation:
        return

    client = LaunchableClient()
    res = client.request("get", session)

    # only check when the status code is 200 not to stop the command
    if res.status_code == 200:
        is_observation_in_recorded_session = res.json().get("isObservation", False)
        if is_observation and not is_observation_in_recorded_session:
            click.echo(
                click.style(
                    "WARNING: --observation flag was ignored. Observation mode can only be enabled for a test session "
                    "during its initial creation. "
                    "Add `--observation` option to the `launchable record session` command instead.",
                    fg='yellow'),
                err=True)



import json
import os
import tempfile
import threading
from http.server import HTTPServer, SimpleHTTPRequestHandler
from pathlib import Path
from unittest import mock

import responses  # type: ignore

from launchable.utils.env_keys import BASE_URL_KEY
from launchable.utils.http_client import get_base_url
from tests.cli_test_case import CliTestCase


# dummy server for exe.jar
# exe.jar calls commits/latest (GET) then commits/collect (POST)
class SuccessCommitHandlerMock(SimpleHTTPRequestHandler):
    def do_GET(self):
        body = []
        self.send_response(200)
        self.send_header("Content-Type", "application/json")
        self.end_headers()
        self.wfile.write(json.dumps(body).encode("utf-8"))

    def do_POST(self):
        self.send_response(200)
        self.send_header("Content-Type", "application/json")
        self.end_headers()


class ErrorCommitHandlerMock(SimpleHTTPRequestHandler):
    def do_GET(self):
        body = []
        self.send_response(200)
        self.send_header("Content-Type", "application/json")
        self.end_headers()
        self.wfile.write(json.dumps(body).encode("utf-8"))

    def do_POST(self):
        body = {"reason": "Internal server error"}
        self.send_response(500)
        self.send_header("Content-Type", "application/json")
        self.end_headers()
        self.wfile.write(json.dumps(body).encode("utf-8"))


class APIErrorTest(CliTestCase):
    test_files_dir = Path(__file__).parent.joinpath('../data/minitest/').resolve()

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_commit(self):
        server = HTTPServer(("", 0), ErrorCommitHandlerMock)
        thread = threading.Thread(None, server.serve_forever)
        thread.start()

        host, port = server.server_address
        endpoint = "http://{}:{}".format(host, port)

        with mock.patch.dict(os.environ, {BASE_URL_KEY: endpoint}):
            result = self.cli("record", "commit", "--source", ".")
            self.assertEqual(result.exit_code, 0)
            self.assertEqual(result.exception, None)

        server.shutdown()
        thread.join()

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_build(self):
        # case: cli catches error
        success_server = HTTPServer(("", 0), SuccessCommitHandlerMock)
        thread = threading.Thread(None, success_server.serve_forever)
        thread.start()

        host, port = success_server.server_address
        endpoint = "http://{}:{}".format(host, port)
        with mock.patch.dict(os.environ, {BASE_URL_KEY: endpoint}):
            responses.add(responses.POST, "{base}/intake/organizations/{org}/workspaces/{ws}/builds".format(
                base=get_base_url(), org=self.organization, ws=self.workspace), status=500)

            result = self.cli("record", "build", "--name", "example")
            self.assertEqual(result.exit_code, 0)
            self.assertEqual(result.exception, None)

        success_server.shutdown()
        thread.join()

        # case: exe.jar catches error
        error_server = HTTPServer(("", 0), ErrorCommitHandlerMock)
        thread = threading.Thread(None, error_server.serve_forever)
        thread.start()

        host, port = error_server.server_address
        endpoint = "http://{}:{}".format(host, port)

        with mock.patch.dict(os.environ, {BASE_URL_KEY: endpoint}):
            responses.add(responses.POST, "{base}/intake/organizations/{org}/workspaces/{ws}/builds".format(
                base=get_base_url(), org=self.organization, ws=self.workspace), status=500)

            result = self.cli("record", "build", "--name", "example")
            self.assertEqual(result.exit_code, 0)
            self.assertEqual(result.exception, None)

        error_server.shutdown()
        thread.join()

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_session(self):
        build = "internal_server_error"
        responses.add(
            responses.POST,
            "{base}/intake/organizations/{org}/workspaces/{ws}/builds/{build}/test_sessions".format(
                base=get_base_url(),
                org=self.organization,
                ws=self.workspace,
                build=build),
            status=500)

        result = self.cli("record", "session", "--build", build)
        self.assertEqual(result.exit_code, 0)

        build = "not_found"
        responses.add(
            responses.POST,
            "{base}/intake/organizations/{org}/workspaces/{ws}/builds/{build}/test_sessions".format(
                base=get_base_url(),
                org=self.organization,
                ws=self.workspace,
                build=build),
            status=404)

        result = self.cli("record", "session", "--build", build)
        self.assertEqual(result.exit_code, 1)

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_subset(self):
        responses.replace(
            responses.POST,
            "{base}/intake/organizations/{org}/workspaces/{ws}/subset".format(
                base=get_base_url(),
                org=self.organization,
                ws=self.workspace),
            status=500)

        subset_file = "example_test.rb"

        with tempfile.NamedTemporaryFile(delete=False) as rest_file:
            result = self.cli(
                "subset",
                "--target",
                "30%",
                "--session",
                self.session,
                "--rest",
                rest_file.name,
                "minitest",
                str(self.test_files_dir) + "/test/**/*.rb",
                mix_stderr=False)

            self.assertEqual(result.exit_code, 0)
            self.assertEqual(len(result.stdout.rstrip().split("\n")), 1)
            self.assertTrue(subset_file in result.stdout)
            self.assertEqual(Path(rest_file.name).read_text(), "")

        responses.replace(responses.POST, "{base}/intake/organizations/{org}/workspaces/{ws}/subset".format(
            base=get_base_url(), org=self.organization, ws=self.workspace), status=404)

        with tempfile.NamedTemporaryFile(delete=False) as rest_file:
            result = self.cli("subset", "--target", "30%", "--session", self.session, "--rest", rest_file.name,
                              "minitest", str(self.test_files_dir) + "/test/**/*.rb", mix_stderr=False)
            self.assertEqual(result.exit_code, 0)

            self.assertEqual(len(result.stdout.rstrip().split("\n")), 1)
            self.assertTrue(subset_file in result.stdout)
            self.assertEqual(Path(rest_file.name).read_text(), "")

    @responses.activate
    @mock.patch.dict(os.environ, {"LAUNCHABLE_TOKEN": CliTestCase.launchable_token})
    def test_record_tests(self):
        responses.replace(
            responses.POST,
            "{base}/intake/organizations/{org}/workspaces/{ws}/builds/{build}/test_sessions/{session_id}/events".format(
                base=get_base_url(),
                org=self.organization,
                ws=self.workspace,
                build=self.build_name,
                session_id=self.session_id),
            json=[], status=500)

        result = self.cli("record", "tests", "--session", self.session, "minitest", str(self.test_files_dir) + "/")
        self.assertEqual(result.exit_code, 0)

        responses.replace(
            responses.POST,
            "{base}/intake/organizations/{org}/workspaces/{ws}/builds/{build}/test_sessions/{session_id}/events".format(
                base=get_base_url(),
                org=self.organization,
                ws=self.workspace,
                build=self.build_name,
                session_id=self.session_id),
            json=[], status=404)

        result = self.cli("record", "tests", "--session", self.session, "minitest", str(self.test_files_dir) + "/")
        self.assertEqual(result.exit_code, 0)



import os
from typing import List

import click

from ...utils.click import KeyValueType
from ...utils.env_keys import REPORT_ERROR_KEY
from ...utils.flavor import normalize_flavors
from ...utils.http_client import LaunchableClient


@click.command()
@click.option(
    '--days',
    'days',
    help='How many days of test sessions in the past to be stat',
    type=int,
    default=7
)
@click.option(
    "--flavor",
    "flavor",
    help='flavors',
    cls=KeyValueType,
    multiple=True,
)
@click.pass_context
def test_sessions(
    context: click.core.Context,
    days: int,
    flavor: List[str] = [],
):
    # We originally wanted to write it like `params: Dict[str, Any]`, but python 3.5 does not support it,
    # so we gave an empty list in the 'flavor' key to give a type hint.
    # If we don't write this, `pip run type` will assume it is Dict[str, int],
    # and the check will not pass.
    params = {'days': days, 'flavor': []}
    flavors = []
    for f in normalize_flavors(flavor):
        flavors.append('%s=%s' % (f[0], f[1]))

    if flavors:
        params['flavor'] = flavors
    else:
        params.pop('flavor', None)

    try:
        client = LaunchableClient(dry_run=context.obj.dry_run)
        res = client.request('get', '/stats/test-sessions', params=params)
        res.raise_for_status()
        click.echo(res.text)

    except Exception as e:
        if os.getenv(REPORT_ERROR_KEY):
            raise e
        else:
            click.echo(e, err=True)
        click.echo(click.style(
            "Warning: the service failed to get stat.", fg='yellow'),
            err=True)
